"""payu.experiment ===============

   Interface to an individual experiment managed by payu

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""

from __future__ import print_function

# Standard Library
import datetime
import errno
import os
import re
import resource
import sys
import shlex
import shutil
import subprocess as sp
import sysconfig
from pathlib import Path

# Extensions
import yaml
from packaging import version

# Local
from payu import envmod
from payu.fsops import mkdir_p, make_symlink, read_config, movetree
from payu.fsops import list_archive_dirs
from payu.fsops import run_script_command
from payu.fsops import needs_subprocess_shell
from payu.schedulers.pbs import get_job_info, pbs_env_init, get_job_id
from payu.models import index as model_index
import payu.profilers
from payu.runlog import Runlog
from payu.manifest import Manifest
from payu.calendar import parse_date_offset
from payu.sync import SyncToRemoteArchive
from payu.metadata import Metadata

# Environment module support on vayu
# TODO: To be removed
core_modules = ['python', 'payu']

# Default payu parameters
default_restart_freq = 5


class Experiment(object):

    def __init__(self, lab, reproduce=False, force=False, metadata_off=False):
        self.lab = lab
        # Check laboratory directories are writable
        self.lab.initialize()

        if not force:
            # check environment for force flag under PBS
            self.force = os.environ.get('PAYU_FORCE', False)
        else:
            self.force = force

        self.start_time = datetime.datetime.now()

        # Initialise experiment metadata - uuid and experiment name
        self.metadata = Metadata(Path(lab.archive_path), disabled=metadata_off)
        self.metadata.setup()

        # TODO: replace with dict, check versions via key-value pairs
        self.modules = set()

        # TODO: __init__ should not be a config dumping ground!
        self.config = read_config()

        # Payu experiment type
        self.debug = self.config.get('debug', False)
        self.postscript = self.config.get('postscript')
        self.repeat_run = self.config.get('repeat', False)

        # Configuration
        self.expand_shell_vars = True   # TODO: configurable

        # Model run time
        self.runtime = None
        if ('calendar' in self.config and
                'runtime' in self.config['calendar']):
            self.runtime = self.config['calendar']['runtime']

        # Stacksize
        # NOTE: Possible PBS issue in setting non-unlimited stacksizes
        stacksize = self.config.get('stacksize', 'unlimited')
        self.set_stacksize(stacksize)

        # Initialize the submodels
        self.init_models()

        # TODO: Move to run/collate/sweep?
        self.set_expt_pathnames()
        self.set_counters()

        for model in self.models:
            model.set_input_paths()

        self.set_output_paths()

        # Create metadata file and move to archive
        self.metadata.write_metadata(restart_path=self.prior_restart_path)

        if not reproduce:
            # check environment for reproduce flag under PBS
            reproduce = os.environ.get('PAYU_REPRODUCE', False)

        # Initialize manifest
        self.manifest = Manifest(self.config.get('manifest', {}),
                                 reproduce=reproduce)

        # Miscellaneous configurations
        # TODO: Move this stuff somewhere else
        self.userscripts = self.config.get('userscripts', {})

        self.profilers = []

        init_script = self.userscripts.get('init')
        if init_script:
            self.run_userscript(init_script)

        self.runlog = Runlog(self)

        # XXX: Temporary spot for the payu path
        #      This is horrible; payu/cli.py does this much more safely!
        #      But also does not even store it in os.environ!
        default_payu_bin = os.path.dirname(sys.argv[0])
        payu_bin = os.environ.get('PAYU_PATH', default_payu_bin)

        self.payu_path = os.path.join(payu_bin, 'payu')

        self.run_id = None

        self.user_modules_paths = None

    def init_models(self):

        self.model_name = self.config.get('model')
        assert self.model_name

        model_fields = ['model', 'exe', 'input', 'ncpus', 'npernode', 'build',
                        'mpthreads', 'exe_prefix']

        # XXX: Temporarily adding this to model config...
        model_fields += ['mask']

        # TODO: Rename this to self.submodels
        self.models = []

        submodels = self.config.get('submodels', [])

        submodel_config = dict((f, self.config[f]) for f in model_fields
                               if f in self.config)
        submodel_config['name'] = self.model_name

        submodels.append(submodel_config)

        for m_config in submodels:
            ModelType = model_index[m_config['model']]
            self.models.append(ModelType(self, m_config['name'], m_config))

        # Load the top-level model
        if self.model_name:
            ModelType = model_index[self.model_name]
            model_config = dict((f, self.config[f]) for f in model_fields
                                if f in self.config)
            self.model = ModelType(self, self.model_name, model_config)
            self.model.top_level_model = True
        else:
            self.model = None

    def set_counters(self):
        # Assume that ``set_paths`` has already been called
        assert self.archive_path

        current_counter = os.environ.get('PAYU_CURRENT_RUN')
        if current_counter:
            self.counter = int(current_counter)
        else:
            self.counter = None

        self.n_runs = int(os.environ.get('PAYU_N_RUNS', 1))

        # Initialize counter if unset
        if self.counter is None:
            # Check for restart index
            max_restart_index = self.max_output_index(output_type="restart")
            if max_restart_index is not None:
                self.counter = 1 + max_restart_index
            else:
                # Now look for output directories,
                # as repeat runs do not generate restart files.
                max_output_index = self.max_output_index()
                if max_output_index is not None:
                    self.counter = 1 + max_output_index
                else:
                    self.counter = 0

    def max_output_index(self, output_type="output"):
        """Given a output directory type (output or restart),
        return the maximum index of output directories found"""
        try:
            output_dirs = list_archive_dirs(archive_path=self.archive_path,
                                            dir_type=output_type)
        except EnvironmentError as exc:
            if exc.errno == errno.ENOENT:
                output_dirs = None
            else:
                raise

        if output_dirs and len(output_dirs):
            return int(output_dirs[-1].lstrip(output_type))

    def set_stacksize(self, stacksize):

        if stacksize == 'unlimited':
            stacksize = resource.RLIM_INFINITY
        else:
            assert type(stacksize) is int

        resource.setrlimit(resource.RLIMIT_STACK,
                           (stacksize, resource.RLIM_INFINITY))

    def setup_modules(self):
        """Setup modules and get paths added to $PATH by user-modules"""
        envmod.setup()

        # Get user modules info from config
        user_modulepaths = self.config.get('modules', {}).get('use', [])
        user_modules = self.config.get('modules', {}).get('load', [])

        # Run module use + load commands for user-defined modules, and
        # get a set of paths and loaded modules added by loading the modules
        loaded_mods, paths = envmod.setup_user_modules(user_modules,
                                                       user_modulepaths)
        self.user_modules_paths = paths
        self.loaded_user_modules = [] if loaded_mods is None else loaded_mods

    def load_modules(self):
        # Scheduler
        sched_modname = self.config.get('scheduler', 'pbs')
        self.modules.add(sched_modname)

        # MPI library
        mpi_config = self.config.get('mpi', {})

        # Assign MPI module paths
        mpi_modpath = mpi_config.get('modulepath', None)
        if mpi_modpath:
            envmod.module('use', mpi_modpath)

        mpi_modname = mpi_config.get('module', 'openmpi')
        self.modules.add(mpi_modname)

        # Unload non-essential modules
        loaded_mods = os.environ.get('LOADEDMODULES', '').split(':')

        for mod in loaded_mods:
            if len(mod) > 0:
                print('mod '+mod)
                mod_base = mod.split('/')[0]
                if (mod_base not in core_modules and
                        mod not in self.loaded_user_modules):
                    envmod.module('unload', mod)

        # Now load model-dependent modules
        for mod in self.modules:
            envmod.module('load', mod)

        envmod.module('list')

        for prof in self.profilers:
            prof.load_modules()

        # TODO: Consolidate this profiling stuff
        c_ipm = self.config.get('ipm', False)
        if c_ipm:
            if isinstance(c_ipm, str):
                ipm_mod = os.path.join('ipm', c_ipm)
            else:
                ipm_mod = 'ipm/2.0.2'

            envmod.module('load', ipm_mod)
            os.environ['IPM_LOGDIR'] = self.work_path

        if self.config.get('mpiP', False):
            envmod.module('load', 'mpiP')

        if self.config.get('hpctoolkit', False):
            envmod.module('load', 'hpctoolkit')

        if self.debug:
            envmod.module('load', 'totalview')

    def set_expt_pathnames(self):

        # Local "control" path default used to be applied here,
        # but now done in read_config
        self.control_path = self.config.get('control_path')

        # Experiment name
        self.name = self.metadata.experiment_name

        # Experiment subdirectories
        self.archive_path = os.path.join(self.lab.archive_path, self.name)
        self.work_path = os.path.join(self.lab.work_path, self.name)

        # Symbolic link paths to output
        self.work_sym_path = os.path.join(self.control_path, 'work')
        self.archive_sym_path = os.path.join(self.control_path, 'archive')

        for model in self.models:
            model.set_model_pathnames()
            model.set_local_pathnames()

        # Stream output filenames
        # TODO: per-model output streams?
        self.stdout_fname = self.lab.model_type + '.out'
        self.stderr_fname = self.lab.model_type + '.err'

        self.job_fname = 'job.yaml'
        self.env_fname = 'env.yaml'

        self.output_fnames = (self.stderr_fname,
                              self.stdout_fname,
                              self.job_fname,
                              self.env_fname)

    def set_output_paths(self):

        # Local archive paths

        # Check to see if we've provided a hard coded path -- valid for collate
        dir_path = os.environ.get('PAYU_DIR_PATH')
        if dir_path is not None:
            self.output_path = os.path.normpath(dir_path)
        else:
            output_dir = 'output{0:03}'.format(self.counter)
            self.output_path = os.path.join(self.archive_path, output_dir)

        # TODO: check case counter == 0
        prior_output_dir = 'output{0:03}'.format(self.counter - 1)
        prior_output_path = os.path.join(self.archive_path, prior_output_dir)
        if os.path.exists(prior_output_path):
            self.prior_output_path = prior_output_path
        else:
            self.prior_output_path = None

        # Local restart paths
        restart_dir = 'restart{0:03}'.format(self.counter)
        self.restart_path = os.path.join(self.archive_path, restart_dir)

        # Prior restart path

        # Check if a user restart directory is avaiable
        user_restart_dir = self.config.get('restart')
        if (self.counter == 0 or self.repeat_run) and user_restart_dir:
            # TODO: Some user friendliness needed...
            assert (os.path.isdir(user_restart_dir))
            self.prior_restart_path = user_restart_dir
        else:
            prior_restart_dir = 'restart{0:03}'.format(self.counter - 1)
            prior_restart_path = os.path.join(self.archive_path,
                                              prior_restart_dir)
            if os.path.exists(prior_restart_path) and not self.repeat_run:
                self.prior_restart_path = prior_restart_path
            else:
                self.prior_restart_path = None
                if self.counter > 0 and not self.repeat_run:
                    # TODO: This warning should be replaced with an abort in
                    #       setup
                    print('payu: warning: No restart files found.')

        for model in self.models:
            model.set_model_output_paths()

    def build_model(self):

        self.load_modules()

        for model in self.models:
            model.get_codebase()

        for model in self.models:
            model.build_model()

    def check_payu_version(self):
        """Check current payu version is greater than minimum required
        payu version, if configured"""
        # TODO: Move this function to a setup file if setup is moved to
        # a separate file?
        minimum_version_fieldname = "payu_minimum_version"
        if minimum_version_fieldname not in self.config:
            # Skip version check
            return

        minimum_version = str(self.config[minimum_version_fieldname])
        try:
            # Attempt to parse the version
            parsed_minimum_version = version.parse(minimum_version)
        except version.InvalidVersion:
            raise ValueError(
                "Invalid version in configuration file (config.yaml) for "
                f"'{minimum_version_fieldname}': {minimum_version}"
            )

        # Get the current version of the package
        current_version = payu.__version__

        # Compare versions
        if version.parse(current_version) < parsed_minimum_version:
            raise RuntimeError(
                f"Payu version {current_version} does not meet the configured "
                f"minimum version. A version >= {minimum_version} is "
                "required to run this configuration."
            )

    def setup(self, force_archive=False):
        # Check version
        self.check_payu_version()

        # Confirm that no output path already exists
        if os.path.exists(self.output_path):
            sys.exit('payu: error: Output path already exists: '
                     '{path}.'.format(path=self.output_path))

        # Confirm that no work path already exists
        if os.path.exists(self.work_path):
            if self.force:
                print('payu: work path already exists.\n'
                      '      Sweeping as --force option is True.')
                self.sweep()
            else:
                sys.exit('payu: error: work path already exists: {path}.\n'
                         '             payu sweep and then payu run'
                         .format(path=self.work_path))

        mkdir_p(self.work_path)

        if force_archive:
            mkdir_p(self.archive_path)
            make_symlink(self.archive_path, self.archive_sym_path)

        # Archive the payu config
        # TODO: This just copies the existing config.yaml file, but we should
        #       reconstruct a new file including default values
        config_src = os.path.join(self.control_path, 'config.yaml')
        config_dst = os.path.join(self.work_path)
        shutil.copy(config_src, config_dst)

        # Stripe directory in Lustre
        # TODO: Make this more configurable
        do_stripe = self.config.get('stripedio', False)
        if do_stripe:
            cmd = 'lfs setstripe -c 8 -s 8m {0}'.format(self.work_path)
            sp.check_call(shlex.split(cmd))

        make_symlink(self.work_path, self.work_sym_path)

        # Set up executable paths - first search through paths added by modules
        self.setup_modules()
        for model in self.models:
            model.setup_executable_paths()

        # Set up all file manifests
        self.manifest.setup()

        for model in self.models:
            model.setup()

        # Call the macro-model setup
        if len(self.models) > 1:
            self.model.setup()

        self.manifest.check_manifests()

        # Copy manifests to work directory so they archived on completion
        manifest_path = os.path.join(self.work_path, 'manifests')
        self.manifest.copy_manifests(manifest_path)

        setup_script = self.userscripts.get('setup')
        if setup_script:
            self.run_userscript(setup_script)

        # Profiler setup
        expt_profs = self.config.get('profilers', [])
        if not isinstance(expt_profs, list):
            expt_profs = [expt_profs]

        for prof_name in expt_profs:
            ProfType = payu.profilers.index[prof_name]
            prof = ProfType(self)
            self.profilers.append(prof)

            # Testing
            prof.setup()

        # Check restart pruning for valid configuration values and
        # warns user if more restarts than expected would be pruned
        if self.archiving():
            self.get_restarts_to_prune()

    def run(self, *user_flags):
        self.load_modules()

        f_out = open(self.stdout_fname, 'w')
        f_err = open(self.stderr_fname, 'w')

        # Set MPI environment variables
        env = self.config.get('env')

        # Explicitly check for `None`, in case of an empty `env:` entry
        if env is None:
            env = {}

        for var in env:

            if env[var] is None:
                env_value = ''
            else:
                env_value = str(env[var])

            os.environ[var] = env_value

        mpi_config = self.config.get('mpi', {})
        mpi_runcmd = mpi_config.get('runcmd', 'mpirun')

        if self.config.get('scalasca', False):
            mpi_runcmd = ' '.join(['scalasca -analyze', mpi_runcmd])

        # MPI runtime flags
        mpi_flags = mpi_config.get('flags', [])
        if not mpi_flags:
            # DEPRECATED: confusing and a duplication of flags config
            if 'mpirun' in self.config:
                mpi_flags = self.config.get('mpirun')
                print('payu: warning: mpirun config option is deprecated.'
                      '  Use mpi: flags option instead')
            else:
                mpi_flags = []

        if not isinstance(mpi_flags, list):
            mpi_flags = [mpi_flags]

        # TODO: More uniform support needed here
        if self.config.get('scalasca', False):
            mpi_flags = ['\"{0}\"'.format(f) for f in mpi_flags]

        # XXX: I think this may be broken
        if user_flags:
            mpi_flags.extend(list(user_flags))

        if self.debug:
            mpi_flags.append('--debug')

        mpi_progs = []
        for model in self.models:

            # Skip models without executables (e.g. couplers)
            if not model.exec_path_local:
                continue

            mpi_config = self.config.get('mpi', {})
            mpi_module = mpi_config.get('module', None)

            # Update MPI library module (if not explicitly set)
            # TODO: Check for MPI library mismatch across multiple binaries
            if mpi_module is None and model.required_libs is not None:
                envmod.lib_update(
                    model.required_libs,
                    'libmpi.so'
                )

            model_prog = []

            wdir_arg = '-wdir'
            if self.config.get('scheduler') == 'slurm':
                # Option to set the working directory differs in slurm
                wdir_arg = '--chdir'
            model_prog.append(f'{wdir_arg} {model.work_path}')

            # Append any model-specific MPI flags
            model_flags = model.config.get('mpiflags', [])
            if not isinstance(model_flags, list):
                model_prog.append(model_flags)
            else:
                model_prog.extend(model_flags)

            model_ncpus = model.config.get('ncpus')
            if model_ncpus:
                if self.config.get('scheduler') == 'slurm':
                    model_prog.append('-n {0}'.format(model_ncpus))
                else:
                    # Default to preferred mpirun syntax
                    model_prog.append('-np {0}'.format(model_ncpus))

            model_npernode = model.config.get('npernode')
            # TODO: New Open MPI format?
            if model_npernode:
                if model_npernode % 2 == 0:
                    npernode_flag = ('-map-by ppr:{0}:socket'
                                     ''.format(model_npernode / 2))
                else:
                    npernode_flag = ('-map-by ppr:{0}:node'
                                     ''.format(model_npernode))

                if self.config.get('scalasca', False):
                    npernode_flag = '\"{0}\"'.format(npernode_flag)
                model_prog.append(npernode_flag)

            if self.config.get('hpctoolkit', False):
                os.environ['HPCRUN_EVENT_LIST'] = 'WALLCLOCK@5000'
                model_prog.append('hpcrun')

            for prof in self.profilers:
                if prof.runscript:
                    model_prog.append(prof.runscript)

            model_prog.append(model.exec_prefix)

            # Use the full path to symlinked exec_name in work as some
            # older MPI libraries complained executable was not in PATH
            model_prog.append(os.path.join(model.work_path, model.exec_name))

            mpi_progs.append(' '.join(model_prog))

        cmd = '{runcmd} {flags} {exes}'.format(
            runcmd=mpi_runcmd,
            flags=' '.join(mpi_flags),
            exes=' : '.join(mpi_progs)
        )

        for prof in self.profilers:
            cmd = prof.wrapper(cmd)

        # Expand shell variables inside flags
        if self.expand_shell_vars:
            cmd = os.path.expandvars(cmd)

        # TODO: Consider making this default
        if self.config.get('coredump', False):
            enable_core_dump()

        # Dump out environment
        with open(self.env_fname, 'w') as file:
            file.write(yaml.dump(dict(os.environ), default_flow_style=False))

        self.runlog.create_manifest()
        if self.runlog.enabled:
            self.runlog.commit()

        # NOTE: This may not be necessary, since env seems to be getting
        # correctly updated.  Need to look into this.
        print(cmd)
        if env:
            # TODO: Replace with mpirun -x flag inputs
            proc = sp.Popen(shlex.split(cmd), stdout=f_out, stderr=f_err,
                            env=os.environ.copy())
            proc.wait()
            rc = proc.returncode
        else:
            rc = sp.call(shlex.split(cmd), stdout=f_out, stderr=f_err)

        f_out.close()
        f_err.close()

        self.finish_time = datetime.datetime.now()

        info = get_job_info()

        if info is None:
            # Not being run under PBS, reverse engineer environment
            info = {
                'PAYU_PATH': os.path.dirname(self.payu_path)
            }

        # Add extra information to save to jobinfo
        info.update(
            {
                'PAYU_CONTROL_DIR': self.control_path,
                'PAYU_RUN_ID': self.run_id,
                'PAYU_CURRENT_RUN': self.counter,
                'PAYU_N_RUNS':  self.n_runs,
                'PAYU_JOB_STATUS': rc,
                'PAYU_START_TIME': self.start_time.isoformat(),
                'PAYU_FINISH_TIME': self.finish_time.isoformat(),
                'PAYU_WALLTIME': "{0} s".format(
                    (self.finish_time - self.start_time).total_seconds()
                ),
            }
        )

        # Dump job info
        with open(self.job_fname, 'w') as file:
            file.write(yaml.dump(info, default_flow_style=False))

        # Remove any empty output files (e.g. logs)
        for fname in os.listdir(self.work_path):
            fpath = os.path.join(self.work_path, fname)
            if os.path.getsize(fpath) == 0:
                os.remove(fpath)

        # Clean up any profiling output
        # TODO: Move after `rc` code check?
        for prof in self.profilers:
            prof.postprocess()

        # TODO: Need a model-specific cleanup method call here
        # NOTE: This does not appear to catch hanging jobs killed by PBS
        if rc != 0:
            # Backup logs for failed runs
            error_log_dir = os.path.join(self.archive_path, 'error_logs')
            mkdir_p(error_log_dir)

            # NOTE: This is PBS-specific
            job_id = get_job_id(short=False)

            if job_id == '':
                job_id = str(self.run_id)[:6]

            for fname in self.output_fnames:

                src = os.path.join(self.control_path, fname)

                stem, suffix = os.path.splitext(fname)
                dest = os.path.join(error_log_dir,
                                    ".".join((stem, job_id)) + suffix)

                print(src, dest)

                shutil.copyfile(src, dest)

            # Create the symlink to the logs if it does not exist
            make_symlink(self.archive_path, self.archive_sym_path)

            error_script = self.userscripts.get('error')
            if error_script:
                self.run_userscript(error_script)

            # Terminate payu
            sys.exit('payu: Model exited with error code {0}; aborting.'
                     ''.format(rc))

        # Decrement run counter on successful run
        stop_file_path = os.path.join(self.control_path, 'stop_run')
        if os.path.isfile(stop_file_path):
            assert os.stat(stop_file_path).st_size == 0
            os.remove(stop_file_path)
            print('payu: Stop file detected; terminating resubmission.')
            self.n_runs = 0
        else:
            self.n_runs -= 1

        # Move logs to archive (or delete if empty)
        for f in self.output_fnames:
            f_path = os.path.join(self.control_path, f)
            if os.path.getsize(f_path) == 0:
                os.remove(f_path)
            else:
                shutil.move(f_path, self.work_path)

        run_script = self.userscripts.get('run')
        if run_script:
            self.run_userscript(run_script)

    def archiving(self):
        """
        Determine whether to run archive step based on config.yaml settings.
        Default to True when archive settings are absent.
        """
        archive_config = self.config.get('archive', {})
        return archive_config.get('enable', True)

    def archive(self, force_prune_restarts=False):
        if not self.archiving():
            print('payu: not archiving due to config.yaml setting.')
            return

        # Check there is a work directory, otherwise bail
        if not os.path.exists(self.work_sym_path):
            sys.exit('payu: error: No work directory to archive.')

        mkdir_p(self.archive_path)
        make_symlink(self.archive_path, self.archive_sym_path)

        # Remove work symlink
        if os.path.islink(self.work_sym_path):
            os.remove(self.work_sym_path)

        mkdir_p(self.restart_path)

        for model in self.models:
            model.archive()

        # Postprocess the model suite
        if len(self.models) > 1:
            self.model.archive()

        # Double-check that the run path does not exist
        if os.path.exists(self.output_path):
            sys.exit('payu: error: Output path already exists.')

        movetree(self.work_path, self.output_path)

        # Remove any outdated restart files
        try:
            restarts_to_prune = self.get_restarts_to_prune(
                force=force_prune_restarts)
        except Exception as e:
            print(e)
            print("payu: error: Skipping pruning restarts")
            restarts_to_prune = []

        for restart in restarts_to_prune:
            restart_path = os.path.join(self.archive_path, restart)
            # Only delete real directories; ignore symbolic restart links
            if (os.path.isdir(restart_path) and
                    not os.path.islink(restart_path)):
                shutil.rmtree(restart_path)

        # Ensure dynamic library support for subsequent python calls
        ld_libpaths = os.environ.get('LD_LIBRARY_PATH', None)
        py_libpath = sysconfig.get_config_var('LIBDIR')
        if ld_libpaths is None:
            os.environ['LD_LIBRARY_PATH'] = py_libpath
        elif py_libpath not in ld_libpaths.split(':'):
            os.environ['LD_LIBRARY_PATH'] = f'{py_libpath}:{ld_libpaths}'

        collate_config = self.config.get('collate', {})
        collating = collate_config.get('enable', True)
        if collating:
            cmd = '{python} {payu} collate -i {expt}'.format(
                python=sys.executable,
                payu=self.payu_path,
                expt=self.counter
            )
            sp.check_call(shlex.split(cmd))

        if self.config.get('hpctoolkit', False):
            cmd = '{python} {payu} profile -i {expt}'.format(
                python=sys.executable,
                payu=self.payu_path,
                expt=self.counter
            )
            sp.check_call(shlex.split(cmd))

        archive_script = self.userscripts.get('archive')
        if archive_script:
            self.run_userscript(archive_script)

        # Ensure postprocessing runs if model not collating
        if not collating:
            self.postprocess()

    def collate(self):
        # Setup modules - load user-defined modules
        self.setup_modules()

        for model in self.models:
            model.collate()

    def profile(self):
        for model in self.models:
            model.profile()

    def postprocess(self):
        """Submit any postprocessing scripts or remote syncing if enabled"""

        # First submit postprocessing script
        if self.postscript:
            self.set_userscript_env_vars()
            envmod.setup()
            envmod.module('load', 'pbs')

            cmd = 'qsub {script}'.format(script=self.postscript)

            if needs_subprocess_shell(cmd):
                sp.check_call(cmd, shell=True)
            else:
                sp.check_call(shlex.split(cmd))

        # Submit a sync script if remote syncing is enabled
        sync_config = self.config.get('sync', {})
        syncing = sync_config.get('enable', False)
        if syncing:
            cmd = '{python} {payu} sync'.format(
                python=sys.executable,
                payu=self.payu_path
            )

            if self.postscript:
                print('payu: warning: postscript is configured, so by default '
                      'the lastest outputs will not be synced. To sync the '
                      'latest output, after the postscript job has completed '
                      'run:\n'
                      '    payu sync')
                cmd += f' --sync-ignore-last'

            sp.check_call(shlex.split(cmd))

    def sync(self):
        # RUN any user scripts before syncing archive
        envmod.setup()
        pre_sync_script = self.userscripts.get('sync')
        if pre_sync_script:
            self.run_userscript(pre_sync_script)

        # Run rsync commmands
        SyncToRemoteArchive(self).run()

    def resubmit(self):
        next_run = self.counter + 1
        cmd = '{python} {payu} run -i {start} -n {n}'.format(
            python=sys.executable,
            payu=self.payu_path,
            start=next_run,
            n=self.n_runs
        )
        cmd = shlex.split(cmd)
        sp.call(cmd)

    def set_userscript_env_vars(self):
        """Save information of output directories and current run to
        environment variables, so they can be accessed via user-scripts"""
        os.environ.update(
            {
                'PAYU_CURRENT_OUTPUT_DIR': self.output_path,
                'PAYU_CURRENT_RESTART_DIR': self.restart_path,
                'PAYU_ARCHIVE_DIR': self.archive_path,
                'PAYU_CURRENT_RUN': str(self.counter)
            }
        )

    def run_userscript(self, script_cmd: str):
        """Run a user defined script or subcommand at various stages of the
        payu submissions"""
        self.set_userscript_env_vars()
        run_script_command(script_cmd,
                           control_path=Path(self.control_path))

    def sweep(self, hard_sweep=False):
        # TODO: Fix the IO race conditions!

        # TODO: model outstreams and pbs logs need to be handled separately
        default_job_name = os.path.basename(os.getcwd())
        short_job_name = str(self.config.get('jobname', default_job_name))[:15]

        log_filenames = [short_job_name + '.o', short_job_name + '.e']
        for postfix in ['_c.o', '_c.e', '_p.o', '_p.e', '_s.o', '_s.e']:
            log_filenames.append(short_job_name[:13] + postfix)

        logs = [
            f for f in os.listdir(os.curdir) if os.path.isfile(f) and (
                f.startswith(tuple(log_filenames))
            )
        ]

        pbs_log_path = os.path.join(self.archive_path, 'pbs_logs')
        legacy_pbs_log_path = os.path.join(self.control_path, 'pbs_logs')

        if os.path.isdir(legacy_pbs_log_path):
            # TODO: New path may still exist!
            assert not os.path.isdir(pbs_log_path)
            print('payu: Moving pbs_logs to {0}'.format(pbs_log_path))
            shutil.move(legacy_pbs_log_path, pbs_log_path)
        else:
            mkdir_p(pbs_log_path)

        for f in logs:
            print('Moving log {0}'.format(f))
            shutil.move(f, os.path.join(pbs_log_path, f))

        if hard_sweep:
            if os.path.isdir(self.archive_path):
                print('Removing archive path {0}'.format(self.archive_path))
                cmd = 'rm -rf {0}'.format(self.archive_path)
                cmd = shlex.split(cmd)
                rc = sp.call(cmd)
                assert rc == 0

            if os.path.islink(self.archive_sym_path):
                print('Removing symlink {0}'.format(self.archive_sym_path))
                os.remove(self.archive_sym_path)

        # Remove stdout/err and yaml dumps
        for f in self.output_fnames:
            if os.path.isfile(f):
                os.remove(f)

        if os.path.isdir(self.work_path):
            print('Removing work path {0}'.format(self.work_path))
            cmd = 'rm -rf {0}'.format(self.work_path)
            cmd = shlex.split(cmd)
            rc = sp.call(cmd)
            assert rc == 0

        if os.path.islink(self.work_sym_path):
            print('Removing symlink {0}'.format(self.work_sym_path))
            os.remove(self.work_sym_path)

    def get_restarts_to_prune(self,
                              ignore_intermediate_restarts=False,
                              force=False):
        """Returns a list of restart directories that can be pruned"""
        # Check if archive path exists
        if not os.path.exists(self.archive_path):
            return []

        # Sorted list of restart directories in archive
        restarts = list_archive_dirs(archive_path=self.archive_path,
                                     dir_type='restart')
        restart_indices = {}
        for restart in restarts:
            restart_indices[restart] = int(restart.lstrip('restart'))

        # TODO: Previous logic was to prune all restarts if self.repeat_run
        # Still need to figure out what should happen in this case
        if self.repeat_run:
            return [os.path.join(self.archive_path, restart)
                    for restart in restarts]

        # Use restart_freq to decide what restarts to prune
        restarts_to_prune = []
        intermediate_restarts, previous_intermediate_restarts = [], []
        restart_freq = self.config.get('restart_freq', default_restart_freq)
        if isinstance(restart_freq, int):
            # Using integer frequency to prune restarts
            for restart, restart_index in restart_indices.items():
                if not restart_index % restart_freq == 0:
                    intermediate_restarts.append(restart)
                else:
                    # Add any intermediate restarts to restarts to prune
                    restarts_to_prune.extend(intermediate_restarts)
                    previous_intermediate_restarts = intermediate_restarts
                    intermediate_restarts = []
        else:
            # Using date-based frequency to prune restarts
            try:
                date_offset = parse_date_offset(restart_freq)
            except ValueError as e:
                print('payu: error: Invalid configuration for restart_freq:',
                      restart_freq)
                raise

            next_dt = None
            for restart in restarts:
                # Use model-driver to parse restart directory for a datetime
                restart_path = os.path.join(self.archive_path, restart)
                try:
                    restart_dt = self.model.get_restart_datetime(restart_path)
                except NotImplementedError:
                    print('payu: error: Date-based restart pruning is not '
                          f'implemented for the {self.model.model_type} '
                          'model. To use integer based restart pruning, '
                          'set restart_freq to an integer value.')
                    raise
                except FileNotFoundError as e:
                    print(f'payu: warning: Ignoring {restart} from date-based '
                          f'restart pruning. Error: {e}')
                    continue
                except Exception:
                    print('payu: error: Error parsing restart directory ',
                          f'{restart} for a datetime to prune restarts.')
                    raise

                if (next_dt is not None and restart_dt < next_dt):
                    intermediate_restarts.append(restart)
                else:
                    # Keep the earliest datetime and use last kept datetime
                    # as point of reference when adding the next time interval
                    next_dt = date_offset.add_to_datetime(restart_dt)

                    # Add intermediate restarts to restarts to prune
                    restarts_to_prune.extend(intermediate_restarts)
                    previous_intermediate_restarts = intermediate_restarts
                    intermediate_restarts = []

        if ignore_intermediate_restarts:
            # Return all restarts that'll eventually be pruned
            restarts_to_prune.extend(intermediate_restarts)
            return restarts_to_prune

        if not force:
            # check environment for --force-prune-restarts flag
            force = os.environ.get('PAYU_FORCE_PRUNE_RESTARTS', False)

        # Flag to check whether more restarts than expected will be deleted
        is_unexpected = restarts_to_prune != previous_intermediate_restarts

        # Restart_history override
        restart_history = self.config.get('restart_history', None)
        if restart_history is not None:
            if not isinstance(restart_history, int):
                raise ValueError("payu: error: restart_history is not an "
                                 f"integer value: {restart_history}")

            if len(restarts) > 0:
                max_index = restart_indices[restarts[-1]]
                index_bound = max_index - restart_history

                # Keep restart_history latest restarts, in addition to the
                # permanently saved restarts defined by restart_freq
                restarts_to_prune.extend(intermediate_restarts)
                restarts_to_prune = [res for res in restarts_to_prune
                                     if restart_indices[res] <= index_bound]

                # Expect at most 1 restart to be pruned with restart_history
                is_unexpected = len(restarts_to_prune) > 1

        # Log out warning if more restarts than expected will be deleted
        if not force and is_unexpected:
            config_info = (f'restart pruning frequency of {restart_freq}')
            if restart_history:
                config_info += f' and restart history of {restart_history}'

            print(f'payu: warning: Current {config_info} would result in '
                  'following restarts being pruned: '
                  f'{" ".join(restarts_to_prune)}\n'
                  'If this is expected, use --force-prune-restarts flag at '
                  'next run or archive (if running archive manually), e.g.:\n'
                  '     payu run --force-prune-restarts, or\n'
                  '     payu archive --force-prune-restarts\n'
                  'Otherwise, no restarts will be pruned')

            # Return empty list to prevent restarts being pruned
            restarts_to_prune = []

        return restarts_to_prune


def enable_core_dump():
    # Newer Intel compilers support 'FOR_DUMP_CORE_FILE' while most support
    # 'decfort_dump_flag'.  Setting both for now, but there may be a more
    # platform-independent way to support this.

    # Enable Fortran core dump
    os.environ['FOR_DUMP_CORE_FILE'] = 'TRUE'
    os.environ['decfort_dump_flag'] = 'TRUE'

    # Allow unlimited core dump file sizes
    resource.setrlimit(resource.RLIMIT_CORE,
                       (resource.RLIM_INFINITY, resource.RLIM_INFINITY))



# coding: utf-8

from payu.experiment import Experiment
from payu.laboratory import Laboratory
import payu.subcommands.args as args

title = 'archive'
parameters = {'description': 'Archive model output after run'}

arguments = [args.model, args.config, args.laboratory,
             args.force_prune_restarts]


def runcmd(model_type, config_path, lab_path, force_prune_restarts):

    lab = Laboratory(model_type, config_path, lab_path)
    expt = Experiment(lab)

    expt.archive(force_prune_restarts)


runscript = runcmd



# coding: utf-8

# Standard Library
import argparse
import os

# Local
from payu import cli
from payu.experiment import Experiment
from payu.laboratory import Laboratory
import payu.subcommands.args as args
from payu import fsops

title = 'sync'
parameters = {'description': 'Sync model output to a remote directory'}

arguments = [args.model, args.config, args.laboratory, args.dir_path,
             args.sync_restarts, args.sync_ignore_last]


def runcmd(model_type, config_path, lab_path, dir_path, sync_restarts,
           sync_ignore_last):

    pbs_config = fsops.read_config(config_path)

    pbs_vars = cli.set_env_vars(lab_path=lab_path,
                                dir_path=dir_path,
                                sync_restarts=sync_restarts,
                                sync_ignore_last=sync_ignore_last)

    sync_config = pbs_config.get('sync', {})

    default_ncpus = 1
    default_queue = 'copyq'
    default_mem = '2GB'
    default_walltime = '10:00:00'

    pbs_config['queue'] = sync_config.get('queue', default_queue)

    pbs_config['ncpus'] = sync_config.get('ncpus', default_ncpus)

    pbs_config['mem'] = sync_config.get('mem', default_mem)

    pbs_config['walltime'] = sync_config.get('walltime', default_walltime)

    sync_jobname = sync_config.get('jobname')
    if not sync_jobname:
        pbs_jobname = pbs_config.get('jobname')
        if not pbs_jobname:
            if dir_path and os.path.isdir(dir_path):
                pbs_jobname = os.path.basename(dir_path)
            else:
                pbs_jobname = os.path.basename(os.getcwd())

        sync_jobname = pbs_jobname[:13] + '_s'

    pbs_config['jobname'] = sync_jobname[:15]

    pbs_config['qsub_flags'] = sync_config.get('qsub_flags', '')

    cli.submit_job('payu-sync', pbs_config, pbs_vars)


def runscript():
    parser = argparse.ArgumentParser()
    for arg in arguments:
        parser.add_argument(*arg['flags'], **arg['parameters'])

    run_args = parser.parse_args()

    pbs_vars = cli.set_env_vars(lab_path=run_args.lab_path,
                                dir_path=run_args.dir_path,
                                sync_restarts=run_args.sync_restarts,
                                sync_ignore_last=run_args.sync_ignore_last)

    for var in pbs_vars:
        os.environ[var] = str(pbs_vars[var])

    lab = Laboratory(run_args.model_type,
                     run_args.config_path,
                     run_args.lab_path)
    expt = Experiment(lab)

    expt.sync()



"""payu.models.cable
   ================

   Driver interface to CABLE

   :copyright: Copyright 2021 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""

# Standard Library
import glob
import os
import shutil

# Extensions
import f90nml
import yaml

# Local
from payu.fsops import mkdir_p
from payu.models.model import Model


def _get_forcing_path(variable, year, input_dir, offset=None, repeat=None):
    """Return the met forcing file path for a given variable and year.
    
    Parameters
    ----------
    variable : str
        Variable name.
    year : int
        Year value.
    input_dir : str
        Path to work input directory.
    offset : list of int, optional
        Offset the current simulation year from `offset[0]` to `offset[1]`
        before inferring the met forcing path.
    repeat : list of int, optional
        Constrain the current simulation year between `repeat[0]` and
        `repeat[1]` (inclusive) before inferring the met forcing path. If the
        year is outside the interval, the constrained year repeats over the
        interval.

    Returns
    -------
    path : str
        Path (relative to control directory) to the inferred met forcing file.

    Raises
    ------
    FileNotFoundError
        If unable to infer met forcing path.
    """
    if offset:
        year += offset[1] - offset[0]
    if repeat:
        year = repeat[0] + ((year - repeat[0]) % (repeat[1] - repeat[0] + 1))
    pattern = os.path.join(input_dir, f"*{variable}*{year}*.nc")
    for path in glob.glob(pattern):
        return path
    msg = f"Unable to infer met forcing path for variable {variable} for year {year}."
    raise FileNotFoundError(msg)


class Cable(Model):

    def __init__(self, expt, name, config):
        super(Cable, self).__init__(expt, name, config)

        self.model_type = 'cable'
        self.default_exec = 'cable'

        self.cable_nml_fname = 'cable.nml'

        self.config_files = [
            self.cable_nml_fname,
            'cable_soilparm.nml',
            'pft_params.nml',
        ]

        self.forcing_year_config = 'cable.forcing_year.yaml'
        self.optional_config_files = [self.forcing_year_config]

        self.met_forcing_vars = [
            "Rainf",
            "Snowf",
            "LWdown",
            "SWdown",
            "PSurf",
            "Qair",
            "Tair",
            "Wind",
        ]

    def set_model_pathnames(self):
        super(Cable, self).set_model_pathnames()

        # TODO: Check for path in filename%type
        self.work_input_path = os.path.join(self.work_path, 'INPUT')
        self.work_init_path = self.work_input_path
        # TODO: Check for path in filename%restart_out
        self.work_restart_path = os.path.join(self.work_path, 'RESTART')

        self.restart_calendar_file = self.model_type + '.res.yaml'
        self.restart_calendar_path = os.path.join(self.work_init_path,
                                                  self.restart_calendar_file)

        self.cable_nml_path = os.path.join(self.work_path,
                                           self.cable_nml_fname)

    def setup(self):
        super(Cable, self).setup()

        self.cable_nml = f90nml.read(self.cable_nml_path)
        if self.prior_restart_path:
            with open(self.restart_calendar_path, 'r') as restart_file:
                self.restart_info = yaml.safe_load(restart_file)
        else:
            self.restart_info = {'year': self.cable_nml['cable']['ncciy']}

        year = self.cable_nml['cable']['ncciy'] = self.restart_info['year']

        self.cable_nml['cable']['filename']['restart_in'] = (
            os.path.join('INPUT', 'restart.nc')
        )
        self.cable_nml['cable']['filename']['restart_out'] = (
            os.path.join('RESTART', 'restart.nc')
        )
        self.cable_nml['cable']['output']['restart'] = True

        forcing_year_config_path = os.path.join(self.work_path, self.forcing_year_config)
        if os.path.exists(forcing_year_config_path):
            with open(forcing_year_config_path, 'r') as file:
                conf = yaml.safe_load(file)
                forcing_year_config = conf if conf else {}
            for var in self.met_forcing_vars:
                path = _get_forcing_path(
                    var, year, self.work_input_path, **forcing_year_config
                )
                self.cable_nml["cable"]["gswpfile"][var] = (
                    os.path.relpath(path, start=self.work_path)
                )

        # Write modified namelist file to work dir
        self.cable_nml.write(
            os.path.join(self.work_path, self.cable_nml_fname),
            force=True
        )

    def archive(self, **kwargs):

        # Save model time to restart next run
        with open(os.path.join(self.restart_path,
                  self.restart_calendar_file), 'w') as restart_file:
            restart = {'year': self.restart_info['year'] + 1}
            restart_file.write(yaml.dump(restart, default_flow_style=False))

        super(Cable, self).archive()

        # Archive the restart files
        mkdir_p(self.restart_path)

        restart_files = [f for f in os.listdir(self.work_restart_path)
                         if f.endswith('restart.nc')]

        for f in restart_files:
            f_src = os.path.join(self.work_restart_path, f)
            shutil.move(f_src, self.restart_path)

        os.rmdir(self.work_restart_path)

        # Move all logs into a logs subdir
        log_path = os.path.join(self.work_path, 'logs')
        mkdir_p(log_path)
        log_files = [f for f in os.listdir(self.work_path)
                     if f.startswith('cable_log')]
        for f in log_files:
            f_src = os.path.join(self.work_path, f)
            shutil.move(f_src, log_path)

    def collate(self):
        pass



import os
import copy
import shutil

import pytest

import payu

from test.common import cd
from test.common import tmpdir, ctrldir, labdir, expt_archive_dir
from test.common import config as config_orig
from test.common import write_config
from test.common import make_all_files, make_random_file, write_metadata
from test.common import make_expt_archive_dir

verbose = True

# Global config
config = copy.deepcopy(config_orig)

# Enable metadata
config.pop('metadata')
pytestmark = pytest.mark.filterwarnings(
    "ignore::payu.git_utils.PayuGitWarning")


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    # Should be taken care of by teardown, in case remnants lying around
    try:
        shutil.rmtree(tmpdir)
    except FileNotFoundError:
        pass

    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
        make_all_files()
        write_metadata()
    except Exception as e:
        print(e)

    # Create 5 restarts and outputs
    for dir_type in ['restart', 'output']:
        for i in range(5):
            path = make_expt_archive_dir(type=dir_type, index=i)
            make_random_file(os.path.join(path, f'test-{dir_type}00{i}-file'))


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        shutil.rmtree(tmpdir)
        print('removing tmp')
    except Exception as e:
        print(e)


def setup_sync(additional_config, add_envt_vars=None):
    """Given additional configuration and envt_vars, return initialised
    class used to build/run rsync commands"""
    # Set experiment config
    test_config = copy.deepcopy(config)
    test_config.update(additional_config)
    write_config(test_config)

    # Set up Experiment
    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        experiment = payu.experiment.Experiment(lab, reproduce=False)

    # Set enviroment vars
    if add_envt_vars is not None:
        for var, value in add_envt_vars.items():
            os.environ[var] = value

    return payu.sync.SyncToRemoteArchive(experiment)


def assert_expected_archive_paths(source_paths,
                                  expected_dirs,
                                  expected_protected_dirs):
    """Check given source archive source paths that it includes
    the expected dirs to sync"""
    dirs, protected_dirs = [], []
    for source_path in source_paths:
        path = source_path.path
        assert os.path.dirname(path) == str(expt_archive_dir)

        dir = os.path.basename(path)
        if source_path.protected:
            protected_dirs.append(dir)
        else:
            dirs.append(dir)

    assert dirs == expected_dirs
    assert protected_dirs == expected_protected_dirs


@pytest.mark.parametrize(
    "envt_vars, expected_outputs, expected_protected_outputs",
    [
        (
            {},
            ['output000', 'output001', 'output002', 'output003'], ['output004']
        ),
        (
            {'PAYU_SYNC_IGNORE_LAST': 'True'},
            ['output000', 'output001', 'output002', 'output003'], []
        ),
    ])
def test_add_outputs_to_sync(envt_vars, expected_outputs,
                             expected_protected_outputs):
    sync = setup_sync(additional_config={}, add_envt_vars=envt_vars)

    # Test function
    sync.add_outputs_to_sync()

    # Assert expected outputs and protected outputs are added
    assert_expected_archive_paths(sync.source_paths,
                                  expected_outputs,
                                  expected_protected_outputs)

    # Tidy up test - Remove any added enviroment variables
    for envt_var in envt_vars.keys():
        del os.environ[envt_var]


@pytest.mark.parametrize(
    "add_config, envt_vars, expected_restarts, expected_protected_restarts",
    [
        (
            {
                "sync": {
                    'restarts': True
                },
                "restart_freq": 5
            }, {},
            [], ['restart000']
        ),
        (
            {
                "sync": {
                    'restarts': True
                },
                "restart_freq": 2
            }, {},
            ['restart000', 'restart002'], ['restart004']
        ),
        (
            {
                "sync": {
                    "restarts": True
                },
                "restart_freq": 2
            }, {'PAYU_SYNC_IGNORE_LAST': 'True'},
            ['restart000', 'restart002'], []
        ),
        (
            {"restart_freq": 3}, {'PAYU_SYNC_RESTARTS': 'True'},
            ['restart000', 'restart001', 'restart002'],
            ['restart003', 'restart004']
        ),
    ])
def test_restarts_to_sync(add_config, envt_vars,
                          expected_restarts, expected_protected_restarts):
    sync = setup_sync(add_config, envt_vars)

    # Test function
    sync.add_restarts_to_sync()

    # Assert expected restarts and protected restarts are added
    assert_expected_archive_paths(sync.source_paths,
                                  expected_restarts,
                                  expected_protected_restarts)

    # Tidy up test - Remove any added enviroment variables
    for envt_var in envt_vars.keys():
        del os.environ[envt_var]


def test_set_destination_path():
    additional_config = {
        "sync": {
            "url": "test.domain",
            "user": "test-usr",
            "path": "remote/path",
        }}
    sync = setup_sync(additional_config=additional_config)

    # Test destination_path
    sync.set_destination_path()
    assert sync.destination_path == "test-usr@test.domain:remote/path"

    # Test value error raised when path is not set
    sync = setup_sync(additional_config={})
    with pytest.raises(ValueError):
        sync.set_destination_path()


@pytest.mark.parametrize(
    "add_config, expected_excludes",
    [
        (
            {
                "sync": {
                    "exclude": ["iceh.????-??-??.nc", "*-DEPRECATED"]
                },
                "collate": {
                    "enable": True
                }
            }, ("--exclude iceh.????-??-??.nc --exclude *-DEPRECATED"
                " --exclude *.nc.*")
        ),
        (
            {
                "sync": {
                    "exclude_uncollated": False
                },
                "collate": {
                    "enable": True
                }
            }, ""
        ),
        (
            {
                "sync": {
                    "exclude": "*-DEPRECATED"
                },
                "collate": {
                    "enable": False
                }
            }, "--exclude *-DEPRECATED"
        )
    ])
def test_set_excludes_flags(add_config, expected_excludes):
    sync = setup_sync(additional_config=add_config)

    # Test setting excludes
    sync.set_excludes_flags()
    assert sync.excludes == expected_excludes


def test_sync():
    # Add some logs
    pbs_logs_path = os.path.join(expt_archive_dir, 'pbs_logs')
    os.makedirs(pbs_logs_path)
    log_filename = 'test_s.e1234'
    test_log_content = 'Test log file content'
    with open(os.path.join(pbs_logs_path, log_filename), 'w') as f:
        f.write(test_log_content)

    # Add nested directories to output000
    nested_output_dirs = os.path.join('output000', 'submodel', 'test_sub-dir')
    nested_output_path = os.path.join(expt_archive_dir, nested_output_dirs)
    os.makedirs(nested_output_path)

    # Add empty uncollated file
    uncollated_file = os.path.join(nested_output_dirs, 'test0.res.nc.0000')
    with open(os.path.join(expt_archive_dir, uncollated_file), 'w'):
        pass

    # Add empty collated file
    collated_file = os.path.join(nested_output_dirs, 'test1.res.nc')
    with open(os.path.join(expt_archive_dir, collated_file), 'w'):
        pass

    # Remote archive path
    remote_archive = tmpdir / 'remote'

    additional_config = {
        "sync": {
            "path": str(remote_archive),
            "runlog": False
        }
    }
    sync = setup_sync(additional_config)

    # Function to test
    sync.run()

    expected_dirs_synced = {'output000', 'output001', 'output002',
                            'output003', 'output004',
                            'pbs_logs', 'metadata.yaml'}

    # Test output is moved to remote dir
    assert set(os.listdir(remote_archive)) == expected_dirs_synced

    # Test inner log files are copied
    remote_log_path = os.path.join(remote_archive, 'pbs_logs', log_filename)
    assert os.path.exists(remote_log_path)

    with open(remote_log_path, 'r') as f:
        assert test_log_content == f.read()

    # Check nested output dirs are synced
    assert os.path.exists(os.path.join(remote_archive, nested_output_dirs))

    # Check that uncollated files are not synced by default
    assert not os.path.exists(os.path.join(remote_archive, uncollated_file))
    assert os.path.exists(os.path.join(remote_archive, collated_file))

    # Check synced file still exist locally
    local_archive_dirs = os.listdir(expt_archive_dir)
    for dir in expected_dirs_synced:
        assert dir in local_archive_dirs

    # Test sync with remove synced files locally flag
    additional_config['sync']['remove_local_files'] = True
    sync = setup_sync(additional_config)
    sync.run()

    # Check synced files are removed from local archive
    # Except for the protected paths (last output in this case)
    for output in ['output000', 'output001', 'output002', 'output003']:
        file_path = os.path.join(expt_archive_dir, dir, f'test-{output}-file')
        assert not os.path.exists(file_path)

    last_output_path = os.path.join(expt_archive_dir, 'output004')
    last_output_file = os.path.join(last_output_path, f'test-output004-file')
    assert os.path.exists(last_output_file)

    # Test sync with remove synced dirs flag as well
    additional_config['sync']['remove_local_dirs'] = True
    sync = setup_sync(additional_config)
    sync.run()

    # Assert synced output dirs removed (except for the last output)
    local_archive_dirs = os.listdir(expt_archive_dir)
    for output in ['output000', 'output001', 'output002', 'output003']:
        assert output not in local_archive_dirs
    assert 'output004' in local_archive_dirs



"""payu.models.nemo
   ================

   Driver interface to NEMO

   :copyright: Copyright 2016 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""

import os
import shutil

from payu.fsops import mkdir_p
from payu.models.model import Model


class Nemo(Model):

    def __init__(self, expt, name, config):
        super(Nemo, self).__init__(expt, name, config)

        self.model_type = 'nemo'
        self.default_exec = 'opa'

        self.config_files = [
            'namelist',
            'namelist_ice',
        ]

    def archive(self, **kwargs):

        # Remove symbolic links
        for f in os.listdir(self.work_input_path):
            f_path = os.path.join(self.work_input_path, f)
            if os.path.islink(f_path):
                os.remove(f_path)

        # Archive the restart files
        print("restart path", self.restart_path)
        mkdir_p(self.restart_path)

        restart_files = [f for f in os.listdir(self.work_path)
                         if f.endswith('.dimg')]

        for f in restart_files:
            f_src = os.path.join(self.work_path, f)
            shutil.move(f_src, self.restart_path)

    def collate(self):
        pass



import copy
import shutil

import pytest

import payu

from test.common import cd
from test.common import tmpdir, ctrldir, labdir, expt_archive_dir
from test.common import config as config_orig
from test.common import write_config
from test.common import make_all_files
from test.common import remove_expt_archive_dirs, make_expt_archive_dir
from test.models.test_mom_mixin import make_ocean_restart_dir

verbose = True

# Global config
config = copy.deepcopy(config_orig)


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    # Should be taken care of by teardown, in case remnants lying around
    try:
        shutil.rmtree(tmpdir)
    except FileNotFoundError:
        pass

    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
        make_all_files()
    except Exception as e:
        print(e)


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        shutil.rmtree(tmpdir)
        print('removing tmp')
    except Exception as e:
        print(e)


@pytest.fixture(autouse=True)
def teardown():
    # Run test
    yield

    # Remove any created restart files
    remove_expt_archive_dirs(type='restart')

    # Remove experiment archive
    expt_archive_dir.rmdir()


def create_test_2Y_1_month_frequency_restarts():
    """Create 2 years + 1 month worth of mom restarts directories
    with 1 month runtimes - starting from 1900/02/01 to 1902/02/01
    e.g (run_date, restart_directory)
    (1900/02/01, restart000)
    (1900/03/01, restart001)
     ...
    (1902/02/01, restart024)"""
    restart_dts = []
    for year in [1900, 1901, 1902]:
        for month in range(1, 13):
            if (year == 1900 and month == 1) or (year == 1902 and month > 2):
                # Ignore the first date and dates from 1902/03/01 onwards
                continue
            restart_dts.append(f"{year}-{month}-01 00:00:00")

    for index, run_dt in enumerate(restart_dts):
        make_ocean_restart_dir(start_dt="1900-01-01 00:00:00",
                               run_dt=run_dt,
                               calendar=4,
                               restart_index=index,
                               additional_path='ocean')


def write_test_config(restart_freq=None, restart_history=None):
    test_config = copy.deepcopy(config)
    test_config['model'] = 'access-om2'
    test_config['submodels'] = [
        {'name': 'atmosphere', 'model': 'yatm'},
        {'name': 'ocean', 'model': 'mom'}
    ]
    if restart_freq:
        test_config['restart_freq'] = restart_freq
    if restart_history:
        test_config['restart_history'] = restart_history

    write_config(test_config)


@pytest.mark.parametrize(
    "restart_freq, restart_history, expected_pruned_restarts_indices",
    [
        ("1MS", None, []),
        ("2MS", None, [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23]),
        ("2MS", 5, [1, 3, 5, 7, 9, 11, 13, 15, 17, 19]),
        ("12MS", None,
         [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
          13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]),
        ("1YS", None,
         [1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
          12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]),
        (1, 1, []),
        (5, 3, [1, 2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14, 16, 17, 18, 19, 21]),
        (5, 7, [1, 2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14, 16, 17]),
        (5, None, [1, 2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14, 16, 17, 18, 19])
    ])
def test_force_prune_restarts(restart_freq,
                              restart_history,
                              expected_pruned_restarts_indices):
    # Test --force-prune-restarts with varying restart_freq and restart_history

    # Create restart files 1900/02/01-restart000 to 1902/02/01-restart024
    create_test_2Y_1_month_frequency_restarts()

    # Set up config
    write_test_config(restart_freq, restart_history)

    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        expt = payu.experiment.Experiment(lab, reproduce=False)

        # Function to test
        restarts_to_prune = expt.get_restarts_to_prune(force=True)

    # Extract out index
    restarts_to_prune_indices = [
        int(restart.lstrip('restart')) for restart in restarts_to_prune
    ]

    assert restarts_to_prune_indices == expected_pruned_restarts_indices


@pytest.mark.parametrize(
    "restarts, restart_freq, restart_history, expected_restart_indices",
    [
        ([
            (0, "1901-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00"),
            (5, "1906-01-01 00:00:00"),
            (6, "1907-01-01 00:00:00")
        ], "3YS", None, [4, 5]),
        ([
            (0, "1901-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00"),
            (5, "1906-01-01 00:00:00"),
            (6, "1907-01-01 00:00:00")
        ], "3YS", 2, [4]),
        ([
            (0, "1901-01-01 00:00:00"),
            (1, "1902-01-01 00:00:00"),
            (2, "1903-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00")
        ], "2YS", 1, []),
        ([
            (0, "1901-01-01 00:00:00"),
            (1, "1902-01-01 00:00:00"),
            (2, "1903-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00")
        ], "2YS", None, []),
        ([
            (0, "1901-01-01 00:00:00"),
            (2, "1903-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
        ], 2, None, []),
        ([
            (0, "1901-01-01 00:00:00"),
            (2, "1903-01-01 00:00:00"),
            (3, "1904-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00"),
        ], 2, None, [3]),
        ([
            (2, "1903-01-01 00:00:00"),
            (4, "1905-01-01 00:00:00"),
            (6, "1907-01-01 00:00:00"),
            (8, "1909-01-01 00:00:00"),
        ], 4, None, []),
    ])
def test_prune_restarts(restarts,
                        restart_freq,
                        restart_history,
                        expected_restart_indices):
    # Create restart files
    for index, datetime in restarts:
        make_ocean_restart_dir(start_dt="1900-01-01 00:00:00",
                               run_dt=datetime,
                               calendar=4,
                               restart_index=index,
                               additional_path='ocean')

    # Set up config
    write_test_config(restart_freq, restart_history)

    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        expt = payu.experiment.Experiment(lab, reproduce=False)

        # Function to test - Note: with force=False which is default
        restarts_to_prune = expt.get_restarts_to_prune()

    # Extract out index
    restarts_to_prune_indices = [
        int(restart.lstrip('restart')) for restart in restarts_to_prune
    ]

    assert restarts_to_prune_indices == expected_restart_indices


def test_prune_restarts_ignores_empty_restart_dirs():
    # Test that in date-based restart pruning ignores any empty restart dirs

    write_test_config(restart_freq='1YS')

    # Create restart files
    restart_datetimes = [(4, "1903-01-01 00:00:00"),
                         (5, "1903-06-01 00:00:00"),
                         (6, "1905-01-01 00:00:00")]
    for index, datetime in restart_datetimes:
        make_ocean_restart_dir(start_dt="1900-01-01 00:00:00",
                               run_dt=datetime,
                               calendar=4,
                               restart_index=index,
                               additional_path='ocean')

    # Create empty restart dirs
    make_expt_archive_dir(type='restart', index=0)
    make_expt_archive_dir(type='restart', index=2)

    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        expt = payu.experiment.Experiment(lab, reproduce=False)

        # Function to test
        restarts_to_prune = expt.get_restarts_to_prune()

    # Extract out index
    restarts_to_prune_indices = [
        int(restart.lstrip('restart')) for restart in restarts_to_prune
    ]

    assert restarts_to_prune_indices == [5]


def test_restart_history_with_empty_archive():
    write_test_config(restart_history=3)
    expt_archive_dir.mkdir()

    with cd(ctrldir):
        lab = payu.laboratory.Laboratory(lab_path=str(labdir))
        expt = payu.experiment.Experiment(lab, reproduce=False)

        # Function to test
        restarts_to_prune = expt.get_restarts_to_prune()

    assert restarts_to_prune == []



import copy
import os
import shutil
from datetime import datetime

import pytest
from unittest.mock import patch, Mock
from ruamel.yaml import YAML

from payu.metadata import Metadata, MetadataWarning

from test.common import cd
from test.common import tmpdir, ctrldir, labdir, archive_dir
from test.common import config as config_orig
from test.common import write_config

verbose = True

# Global config - Remove set experiment and metadata config
config = copy.deepcopy(config_orig)
config.pop("experiment")
config.pop("metadata")

pytestmark = pytest.mark.filterwarnings(
    "ignore::payu.git_utils.PayuGitWarning")


def setup_module(module):
    """
    Put any test-wide setup code in here, e.g. creating test files
    """
    if verbose:
        print("setup_module      module:%s" % module.__name__)

    try:
        tmpdir.mkdir()
    except Exception as e:
        print(e)


def teardown_module(module):
    """
    Put any test-wide teardown code in here, e.g. removing test outputs
    """
    if verbose:
        print("teardown_module   module:%s" % module.__name__)

    try:
        shutil.rmtree(tmpdir)
        print('removing tmp')
    except Exception as e:
        print(e)


def mocked_get_git_user_info(config_key):
    if config_key == 'name':
        return 'mockUser'
    elif config_key == 'email':
        return 'mock@email.com'
    else:
        return None


@pytest.fixture(autouse=True)
def setup_and_teardown():
    try:
        ctrldir.mkdir()
        labdir.mkdir()
    except Exception as e:
        print(e)

    yield

    try:
        shutil.rmtree(ctrldir)
        shutil.rmtree(labdir)
    except Exception as e:
        print(e)


@patch("payu.metadata.GitRepository")
@pytest.mark.parametrize(
    "uuid, experiment_name, previous_metadata, expected_metadata",
    [
        # Test new metadata file created
        (
            "b1f3ce3d-99da-40e4-849a-c8b352948a31",
            "ctrl-branch-b1f3ce3d",
            None,
            {
                "experiment_uuid": "b1f3ce3d-99da-40e4-849a-c8b352948a31",
                "created": '2000-01-01',
                "name": "ctrl-branch-b1f3ce3d",
                "model": "TEST-MODEL",
                "url": "mockUrl",
                "contact": "mockUser",
                "email": "mock@email.com"
            }
        ),
        # Test metadata file updated when new UUID
        (
            "7b90f37c-4619-44f9-a439-f76fdf6ae2bd",
            "Control-Branch-7b90f37c",
            {
                "experiment_uuid": "b3298c7f-01f6-4f0a-be32-ce5d2cfb9a04",
                "contact": "Add your name here",
                "email": "Add your email address here",
                "description": "Add description here",
            },
            {
                "experiment_uuid": "7b90f37c-4619-44f9-a439-f76fdf6ae2bd",
                "description": "Add description here",
                "created": '2000-01-01',
                "name": "Control-Branch-7b90f37c",
                "model": "TEST-MODEL",
                "url": "mockUrl",
                "contact": "mockUser",
                "email": "mock@email.com"
            }
        ),
        # Test extra fields not added with legacy experiments
        (
            "7b90f37c-4619-44f9-a439-f76fdf6ae2bd",
            "ctrl",
            {
                "experiment_uuid": "0f49f2a0-f45e-4c0b-a3b6-4b0bf21f2b75",
                "name": "UserDefinedExperimentName",
                "contact": "TestUser",
                "email": "Test@email.com"
            },
            {
                "experiment_uuid": "7b90f37c-4619-44f9-a439-f76fdf6ae2bd",
                "name": "UserDefinedExperimentName",
                "contact": "TestUser",
                "email": "Test@email.com"
            }
        ),
    ]
)
def test_update_file(mock_repo, uuid, experiment_name,
                     previous_metadata, expected_metadata):
    # Create pre-existing metadata file
    metadata_path = ctrldir / 'metadata.yaml'
    yaml = YAML()
    if previous_metadata is not None:
        with open(metadata_path, 'w') as file:
            yaml.dump(previous_metadata, file)

    # Add mock git values
    mock_repo.return_value.get_origin_url.return_value = "mockUrl"
    mock_repo.return_value.get_user_info.side_effect = mocked_get_git_user_info

    # Setup config
    test_config = config.copy()
    test_config['model'] = "test-model"
    write_config(test_config)

    # Initialise Metadata
    with cd(ctrldir):
        metadata = Metadata(archive_dir)
    metadata.uuid = uuid
    metadata.experiment_name = experiment_name

    # Mock datetime (for created date)
    with patch('payu.metadata.datetime') as mock_date:
        mock_date.now.return_value = datetime(2000, 1, 1)

        # Function to test
        metadata.update_file()

    assert metadata_path.exists and metadata_path.is_file

    with open(metadata_path, 'r') as file:
        metadata = yaml.load(metadata_path)

    assert metadata == expected_metadata


@pytest.mark.parametrize(
    "uuid_exists, keep_uuid, is_new_experiment, "
    "branch_uuid_archive_exists, legacy_archive_exists, catch_warning,"
    "expected_uuid, expected_name",
    [
        # Keep UUID on new experiment - UUID Exists - no archives exist
        (
            True, True, True, False, False, False,
            "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", "ctrl-mock_branch-3d18b3b6"
        ),
        # Keep UUID on new experiment - UUID Exists - legacy archive exists
        (
            True, True, True, False, True, False,
            "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", "ctrl-mock_branch-3d18b3b6"
        ),
        # Keep UUID on not new experiement - UUID Exists -legacy archive exists
        (
            True, True, False, False, True, False,
            "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", "ctrl"
        ),
        # Keep UUID on not new experiment - No UUID - no archives exist
        (
            False, True, True, False, False, False,
            "cb793e91-6168-4ed2-a70c-f6f9ccf1659b", "ctrl-mock_branch-cb793e91"
        ),
        # Experiment setup - No UUID - legacy archive exists
        (
            False, False, False, False, True, True,
            "cb793e91-6168-4ed2-a70c-f6f9ccf1659b", "ctrl"
        ),
        # Experiment setup - No UUID - no archive exists
        (
            False, False, False, False, False, True,
            "cb793e91-6168-4ed2-a70c-f6f9ccf1659b", "ctrl-mock_branch-cb793e91"
        ),
        # Experiment setup - Existing UUID - legacy archive exists
        (
            True, False, False, False, True, False,
            "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", "ctrl"
        ),
        # Experiment setup - Existing UUID - new archive exists
        (
            True, False, False, True, True, False,
            "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", "ctrl-mock_branch-3d18b3b6"
        ),
    ]
)
def test_set_experiment_and_uuid(uuid_exists, keep_uuid, is_new_experiment,
                                 branch_uuid_archive_exists,
                                 legacy_archive_exists, catch_warning,
                                 expected_uuid, expected_name):
    # Setup config and metadata
    write_config(config)
    with cd(ctrldir):
        metadata = Metadata(archive_dir)

    if uuid_exists:
        metadata.uuid = "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136"

    if branch_uuid_archive_exists:
        archive_path = archive_dir / "ctrl-mock_branch-3d18b3b6"
        archive_path.mkdir(parents=True)

    if legacy_archive_exists:
        archive_path = archive_dir / "ctrl"
        archive_path.mkdir(parents=True)

    # Test set UUID and experiment name
    with patch('payu.metadata.GitRepository.get_branch_name') as mock_branch, \
         patch('uuid.uuid4') as mock_uuid:
        mock_branch.return_value = "mock_branch"
        mock_uuid.return_value = "cb793e91-6168-4ed2-a70c-f6f9ccf1659b"

        if catch_warning:
            # Test warning raised
            with pytest.warns(MetadataWarning):
                metadata.setup(is_new_experiment=is_new_experiment,
                               keep_uuid=keep_uuid)
        else:
            metadata.setup(is_new_experiment=is_new_experiment,
                           keep_uuid=keep_uuid)

    assert metadata.experiment_name == expected_name
    assert metadata.uuid == expected_uuid


@pytest.mark.parametrize(
    "archive_metadata_exists, archive_uuid, expected_result",
    [
        # A legacy archive exists, but there's no corresponding metadata
        # in archive
        (
            False, None, True
        ),
        # Archive metadata exists but has no UUID
        (
            True, None, True
        ),
        # Archive metadata exists with same UUID
        (
            True, "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136", True
        ),
        # Archive metadata exists with different UUID
        (
            True, "cb793e91-6168-4ed2-a70c-f6f9ccf1659b", False
        ),
    ]
)
def test_has_archive(archive_metadata_exists, archive_uuid, expected_result):
    # Setup config and metadata
    write_config(config)
    with cd(ctrldir):
        metadata = Metadata(archive_dir)
    metadata.uuid = "3d18b3b6-dd19-49a9-8d9e-c7fa8582f136"

    # Setup archive and it's metadata file
    archive_path = archive_dir / "ctrl"
    archive_path.mkdir(parents=True)

    if archive_metadata_exists:
        archive_metadata = {}

        if archive_uuid is not None:
            archive_metadata["experiment_uuid"] = archive_uuid

        with open(archive_path / 'metadata.yaml', 'w') as file:
            YAML().dump(archive_metadata, file)

    result = metadata.has_archive("ctrl")
    assert result == expected_result


def test_set_configured_experiment_name():
    # Set experiment in config file
    test_config = copy.deepcopy(config)
    test_config['experiment'] = "configuredExperiment"
    write_config(test_config)

    with cd(ctrldir):
        metadata = Metadata(archive_dir)

    # Test configured experiment name is always the set experiment name
    metadata.set_experiment_name()
    assert metadata.experiment_name == "configuredExperiment"

    metadata.set_experiment_name(is_new_experiment=True)
    assert metadata.experiment_name == "configuredExperiment"


@pytest.mark.parametrize(
    "branch, expected_name",
    [(None, "ctrl-cb793e91"),
     ("main", "ctrl-cb793e91"),
     ("master", "ctrl-cb793e91"),
     ("branch", "ctrl-branch-cb793e91")]
)
def test_new_experiment_name(branch, expected_name):
    # Test configured experiment name is the set experiment name
    write_config(config)
    with cd(ctrldir):
        metadata = Metadata(archive_dir)

    metadata.uuid = "cb793e91-6168-4ed2-a70c-f6f9ccf1659b"

    with patch('payu.metadata.GitRepository.get_branch_name') as mock_branch:
        mock_branch.return_value = branch
        experiment = metadata.new_experiment_name()

    assert experiment == expected_name


def test_metadata_enable_false():
    # Set metadata to false in config file
    test_config = copy.deepcopy(config)
    test_config['metadata'] = {
        "enable": False
    }
    write_config(test_config)

    with cd(ctrldir):
        metadata = Metadata(archive_dir)
        metadata.setup()
        metadata.write_metadata()

    # Test UUID kept out of experiment name and metadata file is not written
    assert metadata.experiment_name == "ctrl"
    assert not (ctrldir / "metadata.yaml").exists()


def test_metadata_disable():
    # Set metadata to True in config file
    write_config(config)

    with cd(ctrldir):
        # Pass disabled flag to Metadata initialisation call
        metadata = Metadata(archive_dir, disabled=True)
        metadata.setup()
        metadata.write_metadata()

    # Test UUID kept out of experiment name and metadata file is not written
    assert metadata.experiment_name == "ctrl"
    assert not (ctrldir / "metadata.yaml").exists()


@patch("payu.metadata.GitRepository")
def test_update_file_with_template_metadata_values(mock_repo):
    # Leave out origin URL and git user info
    mock_repo.return_value.get_origin_url.return_value = None
    mock_repo.return_value.get_user_info.return_value = None

    # Setup config
    test_config = config.copy()
    test_config['model'] = "test-model"
    write_config(test_config)

    # Initialise Metadata and UUID and experiment name
    with cd(ctrldir):
        metadata = Metadata(archive_dir)
    metadata.experiment_name = "ctrldir-branch-cb793e91"
    metadata.uuid = "cb793e91-6168-4ed2-a70c-f6f9ccf1659"

    with patch('requests.get') as mock_get:
        # Mock request for json schema
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "The name of the experiment (string)"
                },
                "experiment_uuid": {
                    "type": "string",
                    "format": "uuid",
                    "description": "Unique uuid for the experiment (string)"
                },
                "description": {
                    "type": "string",
                    "description": ("Short description of the experiment "
                                    "(string, < 150 char)")
                },
                "long_description": {
                    "type": "string",
                    "description": ("Long description of the experiment "
                                    "(string)")
                },
                "model": {
                    "type": "array",
                    "items": {"type": ["string", "null"]},
                    "description": ("The name(s) of the model(s) used in the"
                                    " experiment (string)")
                },
            },
            "required": [
                "name",
                "experiment_uuid",
                "description",
                "long_description"
            ]
        }
        mock_get.return_value = mock_response

        # Mock datetime (for created date)
        with patch('payu.metadata.datetime') as mock_date:
            mock_date.now.return_value = datetime(2000, 1, 1)

            # Test function
            metadata.update_file(set_template_values=True)

    # Expect commented template values for non-null fields
    expected_metadata = """experiment_uuid: cb793e91-6168-4ed2-a70c-f6f9ccf1659
created: '2000-01-01'
name: ctrldir-branch-cb793e91
model: TEST-MODEL
description:  # Short description of the experiment (string, < 150 char)
long_description: # Long description of the experiment (string)
"""
    assert (ctrldir / 'metadata.yaml').read_text() == expected_metadata



# coding: utf-8
"""
The payu interface for the ACCESSOM2 coupled model
-------------------------------------------------------------------------------
Contact: Marshall Ward <marshall.ward@anu.edu.au>
-------------------------------------------------------------------------------
Distributed as part of Payu, Copyright 2011 Marshall Ward
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""
from __future__ import print_function

import os
import shutil

from payu.models.model import Model
from payu.models.mom import get_restart_datetime_using_mom_submodel


class AccessOm2(Model):

    def __init__(self, expt, name, config):
        super(AccessOm2, self).__init__(expt, name, config)

        self.model_type = 'access-om2'
        self.config_files = ['accessom2.nml', 'namcouple']

    def setup(self):
        # HACK: only call setup if this is called as a submodule
        if not self.top_level_model:
            super(AccessOm2, self).setup()

    def set_model_pathnames(self):
        super(AccessOm2, self).set_model_pathnames()

        # HACK: overwrite what super has done with these.
        self.work_path = self.expt.work_path
        self.control_path = self.expt.control_path

        self.work_input_path = self.work_path
        self.work_restart_path = self.work_path
        self.work_output_path = self.work_path
        self.work_init_path = self.work_path

        self.work_input_path = os.path.join(self.work_path, 'INPUT')
        self.work_restart_path = os.path.join(self.work_path, 'RESTART')

    def set_input_paths(self):
        super(AccessOm2, self).set_input_paths()

        # HACK: overwrite what super has done with this.
        input_dirs = self.expt.config.get('input')

    def set_model_output_paths(self):
        super(AccessOm2, self).set_model_output_paths()

        # HACK: overwrite what super has done with these.
        self.output_path = self.expt.output_path
        self.restart_path = self.expt.restart_path

        self.prior_output_path = self.expt.prior_output_path
        self.prior_restart_path = self.expt.prior_restart_path

    def archive(self):

        # HACK: only called when this is a submodule
        if not self.top_level_model:
            shutil.rmtree(self.work_input_path)
            for f in os.listdir(self.work_restart_path):
                shutil.move(os.path.join(self.work_restart_path, f),
                            self.restart_path)
            os.rmdir(self.work_restart_path)
        else:
            cice5 = None
            mom = None
            for model in self.expt.models:
                if model.model_type == 'cice5':
                    cice5 = model
                elif model.model_type == 'mom':
                    mom = model

            # Copy restart from ocean into ice area.
            if cice5 is not None and mom is not None:
                o2i_src = os.path.join(mom.work_path, 'o2i.nc')
                o2i_dst = os.path.join(cice5.restart_path, 'o2i.nc')
                shutil.copy2(o2i_src, o2i_dst)

    def collate(self):
        pass

    def get_restart_datetime(self, restart_path):
        """Given a restart path, parse the restart files and
        return a cftime datetime (for date-based restart pruning)"""
       # Use mom submodel to determine restart datetimes
        return get_restart_datetime_using_mom_submodel(
            model=self, 
            restart_path=restart_path
        )


"""payu.cli
   ========

   Command line interface tools

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details
"""

import argparse
import sysconfig
import importlib
import os
import pkgutil
import shlex
import subprocess
import sys
import warnings

import payu
import payu.envmod as envmod
from payu.fsops import is_conda
from payu.models import index as supported_models
from payu.schedulers import index as scheduler_index
import payu.subcommands

# Default configuration
DEFAULT_CONFIG = 'config.yaml'

# Force warnings.warn() to omit the source code line in the message
formatwarning_orig = warnings.formatwarning
warnings.formatwarning = (
    lambda message, category, filename, lineno, line=None: (
        formatwarning_orig(message, category, filename, lineno, line='')
    )
)


def parse():
    """Parse the command line inputs and execute the subcommand."""
    parser = generate_parser()

    # Display help if no arguments are provided
    if len(sys.argv) == 1:
        parser.print_help()
    else:
        args = vars(parser.parse_args())
        run_cmd = args.pop('run_cmd')
        run_cmd(**args)


def generate_parser():
    """Parse the command line inputs generate and return parser."""

    # Build the list of subcommand modules
    modnames = [mod for (_, mod, _)
                in pkgutil.iter_modules(payu.subcommands.__path__,
                                        prefix=payu.subcommands.__name__ + '.')
                if mod.endswith('_cmd')]

    subcmds = [importlib.import_module(mod) for mod in modnames]

    # Construct the subcommand parser
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version',
                        version='payu {0}'.format(payu.__version__))

    subparsers = parser.add_subparsers()

    for cmd in subcmds:
        cmd_parser = subparsers.add_parser(cmd.title, **cmd.parameters)
        cmd_parser.set_defaults(run_cmd=cmd.runcmd)

        for arg in cmd.arguments:
            cmd_parser.add_argument(*arg['flags'], **arg['parameters'])

    return parser


def get_model_type(model_type, config):
    """Determine and validate the active model type."""

    # If no model type is given, then check the config file
    if not model_type:
        model_type = config.get('model')

    # If there is still no model type, try the parent directory
    if not model_type:
        model_type = os.path.basename(os.path.abspath(os.pardir))
        print('payu: warning: Assuming model is {0} based on parent directory '
              'name.'.format(model_type))

    if model_type not in supported_models:
        print('payu: error: Unknown model {0}'.format(model_type))
        sys.exit(-1)


def set_env_vars(init_run=None, n_runs=None, lab_path=None, dir_path=None,
                 reproduce=False, force=False, force_prune_restarts=False,
                 sync_restarts=False, sync_ignore_last=False):
    """Construct the environment variables used by payu for resubmissions."""
    payu_env_vars = {}

    if not is_conda():
        # Setup Python dynamic library link
        lib_paths = sysconfig.get_config_vars('LIBDIR')
        payu_env_vars['LD_LIBRARY_PATH'] = ':'.join(lib_paths)

    if 'PYTHONPATH' in os.environ:
        payu_env_vars['PYTHONPATH'] = os.environ['PYTHONPATH']

    # Set (or import) the path to the PAYU scripts (PAYU_PATH)
    # NOTE: We may be able to use sys.path[0] here.
    payu_binpath = os.environ.get('PAYU_PATH')

    if not payu_binpath or not os.path.isdir(payu_binpath):
        payu_binpath = os.path.dirname(sys.argv[0])

    payu_env_vars['PAYU_PATH'] = payu_binpath

    # Set the run counters
    if init_run:
        init_run = int(init_run)
        assert init_run >= 0
        payu_env_vars['PAYU_CURRENT_RUN'] = init_run

    if n_runs:
        n_runs = int(n_runs)
        assert n_runs > 0
        payu_env_vars['PAYU_N_RUNS'] = n_runs

    # Import explicit project paths
    if lab_path:
        payu_env_vars['PAYU_LAB_PATH'] = os.path.normpath(lab_path)

    if dir_path:
        payu_env_vars['PAYU_DIR_PATH'] = os.path.normpath(dir_path)

    if reproduce:
        payu_env_vars['PAYU_REPRODUCE'] = reproduce

    if force:
        payu_env_vars['PAYU_FORCE'] = force

    if sync_restarts:
        payu_env_vars['PAYU_SYNC_RESTARTS'] = sync_restarts

    if sync_ignore_last:
        payu_env_vars['PAYU_SYNC_IGNORE_LAST'] = sync_ignore_last

    if force_prune_restarts:
        payu_env_vars['PAYU_FORCE_PRUNE_RESTARTS'] = force_prune_restarts

    # Pass through important module related environment variables
    module_env_vars = ['MODULESHOME', 'MODULES_CMD', 'MODULEPATH', 'MODULEV']
    for var in module_env_vars:
        if var in os.environ:
            payu_env_vars[var] = os.environ[var]

    return payu_env_vars


def submit_job(script, config, vars=None):
    """Submit a userscript the scheduler."""

    # TODO: Temporary stub to replicate the old approach
    sched_name = config.get('scheduler', 'pbs')
    sched_type = scheduler_index[sched_name]
    sched = sched_type()
    cmd = sched.submit(script, config, vars)
    print(cmd)

    subprocess.check_call(shlex.split(cmd))



# coding: utf-8
"""
The payu interface for the ACCESS coupled climate model
-------------------------------------------------------------------------------
Contact: Marshall Ward <marshall.ward@anu.edu.au>
-------------------------------------------------------------------------------
Distributed as part of Payu, Copyright 2011 Marshall Ward
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""
from __future__ import print_function

# Standard Library
import errno
import os
import re
import shutil
import sys

# Extensions
import f90nml

# Local
from payu.fsops import make_symlink
from payu.models.model import Model
from payu.models.mom import get_restart_datetime_using_mom_submodel
import payu.calendar as cal


class Access(Model):

    def __init__(self, expt, name, config):
        super(Access, self).__init__(expt, name, config)

        self.model_type = 'access'

        for model in self.expt.models:
            if model.model_type == 'cice' or model.model_type == 'cice5':
                model.config_files = ['cice_in.nml',
                                      'input_ice.nml']
                model.optional_config_files = ['input_ice_gfdl.nml',
                                               'input_ice_monin.nml']

                model.ice_nml_fname = 'cice_in.nml'

                model.access_restarts = ['mice.nc']
                model.copy_restarts = True

                model.set_timestep = model.set_access_timestep

            if model.model_type == 'cice5':
                model.access_restarts.extend(['u_star.nc', 'sicemass.nc'])

            if model.model_type == 'cice':
                # The ACCESS build of CICE assumes that restart_dir is 'RESTART'
                model.get_ptr_restart_dir = lambda : '.'

                # Structure of model coupling namelist
                model.cpl_fname = 'input_ice.nml'
                model.cpl_group = 'coupling'
                model.start_date_nml_name = "restart_date.nml"
                # Experiment initialisation date
                model.init_date_key = "init_date"
                # Start date for new run
                model.inidate_key = "inidate"
                # Total time in seconds since initialisation date
                model.runtime0_key = 'runtime0'
                # Simulation length in seconds for new run
                model.runtime_key = "runtime"

            if model.model_type == 'matm':
                # Structure of model coupling namelist
                model.cpl_fname = 'input_atm.nml'
                model.cpl_group = 'coupling'
                model.start_date_nml_name = "restart_date.nml"
                # Experiment initialisation date
                model.init_date_key = "init_date"
                # Start date for new run
                model.inidate_key = "inidate"
                # Total time in seconds since initialisation date
                model.runtime0_key = 'truntime0'
                # Simulation length in seconds for new run
                model.runtime_key = "runtime"


    def setup(self):
        if not self.top_level_model:
            return

        # Keep track of this in order to set the oasis runtime.
        run_runtime = 0

        for model in self.expt.models:

            if model.model_type == 'cice' or model.model_type == 'cice5':

                # Horrible hack to make a link to o2i.nc in the
                # work/ice/RESTART directory
                f_name = 'o2i.nc'
                f_src = os.path.join(model.work_path, f_name)
                f_dst = os.path.join(model.work_restart_path, f_name)

                if os.path.isfile(f_src):
                    make_symlink(f_src, f_dst)

            if model.model_type == 'cice5':

                # Stage the supplemental input files
                if model.prior_restart_path:
                    for f_name in model.access_restarts:
                        f_src = os.path.join(model.prior_restart_path, f_name)
                        f_dst = os.path.join(model.work_input_path, f_name)

                        if os.path.isfile(f_src):
                            make_symlink(f_src, f_dst)

            if model.model_type in ('cice', 'matm'):

                # Update the supplemental OASIS namelists

                # cpl_nml is the coupling namelist copied from the control to
                # work directory.
                cpl_fpath = os.path.join(model.work_path, model.cpl_fname)
                cpl_nml = f90nml.read(cpl_fpath)
                cpl_group = cpl_nml[model.cpl_group]

                # Which calendar are we using, noleap or Gregorian.
                caltype = cpl_group['caltype']

                # Get timing information for the new run.
                if model.prior_restart_path and not self.expt.repeat_run:
                    # Read the start date from the restart date namelist.
                    start_date_fpath = os.path.join(
                        model.prior_restart_path,
                        model.start_date_nml_name
                    )

                    try:
                        start_date_nml = f90nml.read(start_date_fpath)[
                            model.cpl_group]
                    except FileNotFoundError:
                        print(
                            "Missing restart date file for model "
                            f"{model.model_type}",
                            file=sys.stderr
                        )
                        raise

                    # Experiment initialisation date
                    init_date = cal.int_to_date(
                        start_date_nml[model.init_date_key]
                    )

                    # Start date of new run
                    run_start_date = cal.int_to_date(
                        start_date_nml[model.inidate_key]
                    )

                    # run_start_date must be after initialisation date
                    if run_start_date < init_date:
                        msg = (
                            "Restart date 'inidate` in "
                            f"{model.start_date_nml_name} must not be "
                            "before initialisation date `init_date. "
                            "Values provided: \n"
                            f"inidate={start_date_nml[model.inidate_key]}\n"
                            f"init_date={start_date_nml[model.init_date_key]}"
                        )
                        raise ValueError(msg)

                    # Calculate the total number of seconds between the
                    # initialisation and new run start date,
                    # to use for the runtime0 field.
                    previous_runtime = cal.seconds_between_dates(
                        init_date,
                        run_start_date,
                        caltype
                    )

                else:
                    init_date = cal.int_to_date(
                        cpl_group[model.init_date_key]
                    )
                    previous_runtime = 0
                    run_start_date = init_date

                # Get new runtime for this run. We get this from either the
                # 'runtime' part of the payu config, or from the namelist
                if self.expt.runtime:
                    run_runtime = cal.runtime_from_date(
                        run_start_date,
                        self.expt.runtime['years'],
                        self.expt.runtime['months'],
                        self.expt.runtime['days'],
                        self.expt.runtime.get('seconds', 0),
                        caltype)
                else:
                    run_runtime = cpl_group[model.runtime_key]

                # Now write out new run start date and total runtime into the
                # work directory namelist.
                cpl_group[model.init_date_key] = cal.date_to_int(init_date)
                cpl_group[model.inidate_key] = cal.date_to_int(run_start_date)
                cpl_group[model.runtime0_key] = previous_runtime
                cpl_group[model.runtime_key] = int(run_runtime)

                if model.model_type == 'cice':
                    if self.expt.counter and not self.expt.repeat_run:
                        cpl_group['jobnum'] = (
                            1 + self.expt.counter
                        )
                    else:
                        cpl_group['jobnum'] = 1

                nml_work_path = os.path.join(model.work_path, model.cpl_fname)

                # TODO: Does this need to be split into two steps?
                f90nml.write(cpl_nml, nml_work_path + '~')
                shutil.move(nml_work_path + '~', nml_work_path)

            if model.model_type == 'cice':
                if model.prior_restart_path and not self.expt.repeat_run:
                    # Set up and check the cice restart files.
                    model.overwrite_restart_ptr(run_start_date,
                                                previous_runtime,
                                                start_date_fpath)

        # Now change the oasis runtime. This needs to be done after the others.
        for model in self.expt.models:
            if model.model_type == 'oasis':
                namcouple = os.path.join(model.work_path, 'namcouple')

                s = ''
                with open(namcouple, 'r+') as f:
                    s = f.read()
                    m = re.search(r"^[ \t]*\$RUNTIME.*?^[ \t]*(\d+)", s,
                                  re.MULTILINE | re.DOTALL)
                    assert (m is not None)
                    s = s[:m.start(1)] + str(run_runtime) + s[m.end(1):]

                with open(namcouple, 'w') as f:
                    f.write(s)

    def archive(self):
        if not self.top_level_model:
            return

        cice5 = None
        mom = None

        for model in self.expt.models:
            if model.model_type == 'cice':

                # Copy supplemental restart files to RESTART path
                for f_name in model.access_restarts:
                    f_src = os.path.join(model.work_path, f_name)
                    f_dst = os.path.join(model.restart_path, f_name)

                    if os.path.exists(f_src):
                        shutil.move(f_src, f_dst)

                # Copy "cice_in.nml" from work path to restart.
                work_ice_nml_path = os.path.join(
                                        model.work_path,
                                        model.ice_nml_fname
                )
                restart_ice_nml_path = os.path.join(
                                        model.restart_path,
                                        model.ice_nml_fname
                )

                if os.path.exists(work_ice_nml_path):
                    shutil.copy2(work_ice_nml_path, restart_ice_nml_path)

            if model.model_type in ('cice', 'matm'):
                # Write the simulation end date to the restart date
                # namelist.

                # Calculate the end date using information from the work
                # directory coupling namelist.
                work_cpl_fpath = os.path.join(model.work_path, model.cpl_fname)
                work_cpl_nml = f90nml.read(work_cpl_fpath)
                work_cpl_grp = work_cpl_nml[model.cpl_group]

                # Timing information on the completed run.
                exp_init_date_int = work_cpl_grp[model.init_date_key]
                run_start_date_int = work_cpl_grp[model.inidate_key]
                run_runtime = work_cpl_grp[model.runtime_key]
                run_caltype = work_cpl_grp["caltype"]

                # Calculate end date of completed run
                run_end_date = cal.date_plus_seconds(
                    cal.int_to_date(run_start_date_int),
                    run_runtime,
                    run_caltype
                )

                end_date_dict = {
                    model.cpl_group: {
                        model.init_date_key: exp_init_date_int,
                        model.inidate_key: cal.date_to_int(run_end_date)
                    }
                }

                # Write restart date to the restart directory
                end_date_path = os.path.join(model.restart_path,
                                             model.start_date_nml_name)
                f90nml.write(end_date_dict, end_date_path, force=True)

            if model.model_type == 'cice5':
                cice5 = model
            elif model.model_type == 'mom':
                mom = model

        # Copy restart from ocean into ice area.
        if cice5 is not None and mom is not None:
            o2i_src = os.path.join(mom.work_path, 'o2i.nc')
            o2i_dst = os.path.join(cice5.restart_path, 'o2i.nc')
            shutil.copy2(o2i_src, o2i_dst)

    def get_restart_datetime(self, restart_path):
        """Given a restart path, parse the restart files and
        return a cftime datetime (for date-based restart pruning)"""
        return get_restart_datetime_using_mom_submodel(
            model=self, 
            restart_path=restart_path
        )

    def set_model_pathnames(self):
        pass

    def set_local_pathnames(self):
        pass

    def set_input_paths(self):
        pass

    def set_model_output_paths(self):
        pass

    def collate(self):
        pass



import copy
from pathlib import Path
import pytest
import shutil
from unittest.mock import patch

import payu

from .common import cd, make_random_file, get_manifests
from .common import tmpdir, ctrldir, labdir, workdir
from .common import payu_init, payu_setup
from .common import config as config_orig
from .common import write_config
from .common import make_exe, make_inputs

# Config files in the test model driver
CONFIG_FILES = ['data', 'diag', 'input.nml']
OPTIONAL_CONFIG_FILES = ['opt_data']
INPUT_NML_FILENAME = 'input.nml'


@pytest.fixture(autouse=True)
def setup_and_teardown():
    # Create tmp, lab and control directories
    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
    except Exception as e:
        print(e)

    yield

    # Remove tmp directory
    try:
        shutil.rmtree(tmpdir)
    except Exception as e:
        print(e)


def test_init():
    write_config(config_orig)

    # Initialise a payu laboratory
    with cd(ctrldir):
        payu_init(None, None, str(labdir))

    # Check all the correct directories have been created
    for subdir in ['bin', 'input', 'archive', 'codebase']:
        assert((labdir / subdir).is_dir())


def make_config_files():
    """
    Create files required for test model
    """
    for file in CONFIG_FILES:
        make_random_file(ctrldir/file, 29)


def run_payu_setup(config=config_orig, create_inputs=False,
                   create_config_files=True):
    """Helper function to write config.yaml files, make inputs,
    config files and run experiment setup"""
    # Setup files
    write_config(config)
    make_exe()
    if create_inputs:
        make_inputs()
    if create_config_files:
        make_config_files()

    # Initialise a payu laboratory
    with cd(ctrldir):
        payu_init(None, None, str(labdir))

    # Run payu setup
    payu_setup(lab_path=str(labdir))


def test_setup_configuration_files():
    """Test model config_files are copied to work directory,
    and that any symlinks are followed"""
    # Create configuration files
    all_config_files = CONFIG_FILES + OPTIONAL_CONFIG_FILES
    for file in all_config_files:
        if file != INPUT_NML_FILENAME:
            make_random_file(ctrldir / file, 8)

    # For input.nml, create a file symlink in control directory
    input_nml_realpath = tmpdir / INPUT_NML_FILENAME
    input_nml_symlink = ctrldir / INPUT_NML_FILENAME
    make_random_file(input_nml_realpath, 8)
    input_nml_symlink.symlink_to(input_nml_realpath)
    assert input_nml_symlink.is_symlink()

    # Run payu setup
    run_payu_setup(create_inputs=True)

    # Check config files have been copied to work path
    for file in all_config_files + ['config.yaml']:
        filepath = workdir / file
        assert filepath.exists() and filepath.is_file()
        assert not filepath.is_symlink()
        assert filepath.read_bytes() == (ctrldir / file).read_bytes()


@pytest.mark.parametrize(
    "input_path, is_symlink, is_absolute",
    [
        (labdir / 'input' / 'lab_inputs', False, False),
        (ctrldir / 'ctrl_inputs', False, False),
        (tmpdir / 'symlinked_inputs', True, False),
        (tmpdir / 'tmp_inputs', False, True)
    ]
)
def test_setup_inputs(input_path, is_symlink, is_absolute):
    """Test inputs are symlinked to work directory,
    and added in input manifest"""
    # Make inputs
    input_path.mkdir(parents=True, exist_ok=True)
    for i in range(1, 4):
        make_random_file(input_path / f'input_00{i}.bin', 1000**2 + i)

    if is_symlink:
        # Create an input symlink in control directory
        (ctrldir / input_path.name).symlink_to(input_path)

    # Modify config to specify input path
    config = copy.deepcopy(config_orig)
    config['input'] = str(input_path) if is_absolute else input_path.name

    # Run payu setup
    run_payu_setup(config=config, create_config_files=True)

    input_manifest = get_manifests(ctrldir/'manifests')['input.yaml']
    for i in range(1, 4):
        filename = f'input_00{i}.bin'
        work_input = workdir / filename

        # Check file exists and file size is expected
        assert work_input.exists() and work_input.is_symlink()
        assert work_input.stat().st_size == 1000**2 + i

        # Check relative input path is added to manifest
        filepath = str(Path('work') / filename)
        assert filepath in input_manifest

        # Check manifest fullpath
        manifest_fullpath = input_manifest[filepath]['fullpath']
        assert manifest_fullpath == str(input_path / filename)

        # Check fullpath is a resolved path
        assert Path(manifest_fullpath).is_absolute()
        assert not Path(manifest_fullpath).is_symlink()


def test_setup():
    """Test work directory and executable are setup as expected,
    and re-running setup requires a force=True"""
    run_payu_setup(create_inputs=True, create_config_files=True)

    assert workdir.is_symlink and workdir.is_dir()

    # Check executable symlink is in work directory
    bin_exe = labdir / 'bin' / config_orig['exe']
    work_exe = workdir / config_orig['exe']
    assert work_exe.exists() and work_exe.is_symlink()
    assert work_exe.resolve() == bin_exe.resolve()

    # Re-run setup - expect an error
    with pytest.raises(SystemExit,
                       match="work path already exists") as setup_error:
        payu_setup(lab_path=str(labdir), sweep=False, force=False)
    assert setup_error.type == SystemExit

    assert workdir.is_symlink and workdir.is_dir()


@pytest.mark.parametrize(
    "current_version, min_version",
    [
        ("2.0.0", "1.0.0"),
        ("v0.11.2", "v0.11.1"),
        ("1.0.0", "1.0.0"),
        ("1.0.0+4.gabc1234", "1.0.0"),
        ("1.0.0+0.gxyz987.dirty", "1.0.0"),
        ("1.1.5", 1.1)
    ]
)
def test_check_payu_version_pass(current_version, min_version):
    # Mock the payu version
    with patch('payu.__version__', current_version):
        # Avoid running Experiment init method
        with patch.object(payu.experiment.Experiment, '__init__',
                          lambda x: None):
            expt = payu.experiment.Experiment()

            # Mock config.yaml
            expt.config = {
                "payu_minimum_version": min_version
            }
            expt.check_payu_version()


@pytest.mark.parametrize(
    "current_version, min_version",
    [
        ("1.0.0", "2.0.0"),
        ("v0.11", "v0.11.1"),
        ("1.0.0+4.gabc1234", "1.0.1"),
        ("1.0.0+0.gxyz987.dirty", "v1.2"),
    ]
)
def test_check_payu_version_fail(current_version, min_version):
    with patch('payu.__version__', current_version):
        with patch.object(payu.experiment.Experiment, '__init__',
                          lambda x: None):
            expt = payu.experiment.Experiment()

            expt.config = {
                "payu_minimum_version": min_version
            }

            with pytest.raises(RuntimeError):
                expt.check_payu_version()


@pytest.mark.parametrize(
    "current_version", ["1.0.0", "1.0.0+4.gabc1234"]
)
def test_check_payu_version_pass_with_no_minimum_version(current_version):
    with patch('payu.__version__', current_version):
        with patch.object(payu.experiment.Experiment, '__init__',
                          lambda x: None):
            expt = payu.experiment.Experiment()

            # Leave version out of config.yaml
            expt.config = {}

            # Check runs without an error
            expt.check_payu_version()


@pytest.mark.parametrize(
    "minimum_version", ["abcdefg", None]
)
def test_check_payu_version_configured_invalid_version(minimum_version):
    with patch('payu.__version__', "1.0.0"):
        with patch.object(payu.experiment.Experiment, '__init__',
                          lambda x: None):
            expt = payu.experiment.Experiment()

            expt.config = {
                "payu_minimum_version": minimum_version
            }

            with pytest.raises(ValueError):
                expt.check_payu_version()



"""A climate model workflow management tool.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from . import _version
__version__ = _version.get_versions()['version']



"""Driver interface to the FMS model framework.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
:license: Apache License, Version 2.0, see LICENSE for details
"""
from __future__ import print_function

from collections import defaultdict
from pathlib import Path
import multiprocessing
import os
import resource as res
import shlex
import shutil
import subprocess as sp
import sys
from itertools import count
import fnmatch

from payu.models.model import Model
from payu import envmod
from payu.fsops import required_libs

# There is a limit on the number of command line arguments in a forked
# MPI process. This applies only to mppnccombine-fast. The limit is higher
# than this, but mppnccombine-fast is very slow with large numbers of files
MPI_FORK_MAX_FILE_LIMIT = 1000


def cmdthread(cmd, cwd):
    # This is run in a thread, so the GIL of python makes it sensible to
    # capture the output from each process and print it out at the end so
    # it doesn't get scrambled when collates are run in parallel
    output = ''
    returncode = None
    try:
        output = sp.check_output(shlex.split(cmd), cwd=cwd, stderr=sp.STDOUT)
    except sp.CalledProcessError as e:
        output = e.output
        returncode = e.returncode
    return returncode, output


def get_uncollated_files(dir):

    if not os.path.isdir(dir):
        return []

    # Generate collated file list and identify the first tile
    tile_fnames = [f for f in Path(dir).glob('*.nc.*')
                   if f.suffixes[-2] == '.nc' and
                   f.suffixes[-1][1:].isdigit()]

    # Sort numerically according to the number in the suffix and strip off
    # path information
    return [f.name for f
            in sorted(tile_fnames, key=lambda e: int(e.suffixes[-1][1:]))]

def fms_collate(model):
    """
    Collate model output using mppnccombine. This is broken out of the Fms class so that it can
    be used by other models

    Parameters
    ----------
    model: :py:class:`~payu.models.model`
        The payu model class or child thereof
    """
    
    # Set the stacksize to be unlimited
    res.setrlimit(res.RLIMIT_STACK, (res.RLIM_INFINITY, res.RLIM_INFINITY))

    collate_config = model.expt.config.get('collate', {})

    # The mpi flag implies using mppnccombine-fast
    mpi = collate_config.get('mpi', False)

    if mpi:
        # Load mpi modules for collation
        model.expt.load_modules()
        default_exe = 'mppnccombine-fast'
    else:
        default_exe = 'mppnccombine'

    # Locate the FMS collation tool
    # Check config for collate executable
    mppnc_path = collate_config.get('exe')
    if mppnc_path is None:
        for f in os.listdir(model.expt.lab.bin_path):
            if f == default_exe:
                mppnc_path = os.path.join(model.expt.lab.bin_path, f)
                break
    else:
        mppnc_path = model.expand_executable_path(mppnc_path)

    assert mppnc_path, 'No mppnccombine program found'

    # Check config for collate command line options
    collate_flags = collate_config.get('flags')
    if collate_flags is None:
        if mpi:
            collate_flags = '-r'
        else:
            collate_flags = '-n4 -z -m -r'

    if mpi:
        # The output file is the first argument after the flags
        # and mppnccombine-fast uses an explicit -o flag to specify
        # the output
        collate_flags = " ".join([collate_flags, '-o'])
        envmod.lib_update(required_libs(mppnc_path), 'libmpi.so')

    # Import list of collated files to ignore
    collate_ignore = collate_config.get('ignore')
    if collate_ignore is None:
        collate_ignore = []
    elif type(collate_ignore) != list:
        collate_ignore = [collate_ignore]

    # Generate collated file list and identify the first tile
    tile_fnames = {}
    fnames = get_uncollated_files(model.output_path)
    tile_fnames[model.output_path] = fnames

    print(tile_fnames)

    if (collate_config.get('restart', False) and
            model.prior_restart_path is not None):
        # Add uncollated restart files
        fnames = get_uncollated_files(model.prior_restart_path)
        tile_fnames[model.prior_restart_path] = fnames

    # mnc_tiles = defaultdict(list)
    mnc_tiles = defaultdict(defaultdict(list).copy)
    for t_dir in tile_fnames:
        for t_fname in tile_fnames[t_dir]:
            t_base, t_ext = os.path.splitext(t_fname)
            t_ext = t_ext.lstrip('.')

            # Skip any files listed in the ignore list
            if t_base in collate_ignore:
                continue

            mnc_tiles[t_dir][t_base].append(t_fname)

    # print(mnc_tiles)

    if mpi and collate_config.get('glob', True):
        for t_base in mnc_tiles:
            globstr = "{}.*".format(t_base)
            # Try an equivalent glob and check the same files are returned
            mnc_glob = fnmatch.filter(os.listdir(model.output_path),
                                      globstr)
            if mnc_tiles[t_base] == sorted(mnc_glob):
                mnc_tiles[t_base] = [globstr, ]
                print("Note: using globstr ({}) for collating {}"
                      .format(globstr, t_base))
            else:
                print("Warning: cannot use globstr {} to collate {}"
                      .format(globstr, t_base))
                if len(mnc_tiles[t_base]) > MPI_FORK_MAX_FILE_LIMIT:
                    print("Warning: large number of tiles: {} "
                          .format(len(mnc_tiles[t_base])))
                    print("Warning: collation will be slow and may fail")

    cpucount = int(collate_config.get('ncpus',
                   multiprocessing.cpu_count()))

    if mpi:
        # Default to one for mpi
        nprocesses = int(collate_config.get('threads', 1))
    else:
        nprocesses = int(collate_config.get('threads', cpucount))

    ncpusperprocess = int(cpucount/nprocesses)

    if ncpusperprocess == 1 and mpi:
        print("Warning: running collate with mpirun on a single processor")

    pool = multiprocessing.Pool(processes=nprocesses)

    # Collate each tileset into a single file
    results = []
    codes = []
    outputs = []
    for output_path in mnc_tiles:
        for nc_fname in mnc_tiles[output_path]:
            nc_path = os.path.join(output_path, nc_fname)

            # Remove the collated file if it already exists, since it is
            # probably from a failed collation attempt
            # TODO: Validate this somehow
            if os.path.isfile(nc_path):
                os.remove(nc_path)

            cmd = ' '.join([mppnc_path, collate_flags, nc_fname,
                            ' '.join(mnc_tiles[output_path][nc_fname])])
            if mpi:
                cmd = "mpirun -n {} {}".format(ncpusperprocess, cmd)

            print(cmd)
            results.append(
                pool.apply_async(cmdthread, args=(cmd, output_path)))

    pool.close()
    pool.join()

    for result in results:
        rc, op = result.get()
        codes.append(rc)
        outputs.append(op)

    # TODO: Categorise the return codes
    if any(rc is not None for rc in codes):
        for p, rc, op in zip(count(), codes, outputs):
            if rc is not None:
                print('payu: error: Thread {p} crashed with error code '
                      '{rc}.'.format(p=p, rc=rc), file=sys.stderr)
                print(' Error message:', file=sys.stderr)
                print(op.decode(), file=sys.stderr)
        sys.exit(-1)


class Fms(Model):

    def __init__(self, expt, name, config):
        
        # payu initialisation
        super().__init__(expt, name, config)

    def set_model_pathnames(self):

        super().set_model_pathnames()

        # Define local FMS directories
        self.work_restart_path = os.path.join(self.work_path, 'RESTART')
        self.work_input_path = os.path.join(self.work_path, 'INPUT')
        self.work_init_path = self.work_input_path

    def archive(self, **kwargs):
        super().archive()

        # Remove the 'INPUT' path
        shutil.rmtree(self.work_input_path, ignore_errors=True)

        # Archive restart files before processing model output
        if os.path.isdir(self.restart_path):
            os.rmdir(self.restart_path)

        shutil.move(self.work_restart_path, self.restart_path)

    def collate(self):
        fms_collate(self)
